import os
from pathlib import Path
from typing import Type
from urllib.request import urlretrieve

import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any, Literal
from jaxonlayers.functions import kaiming_init_conv2d
from jaxonlayers.layers import BatchNorm
from jaxtyping import Array, Float, PRNGKeyArray
from statedict2pytree import (
    convert,
    move_running_fields_to_the_end,
    pytree_to_fields,
    state_dict_to_fields,
)

from jaxonmodels.functions import (
    default_floating_dtype,
    dtype_to_str,
)

_MODELS = {
    "resnet18_IMAGENET1K_V1": "https://download.pytorch.org/models/resnet18-f37072fd.pth",
    "resnet34_IMAGENET1K_V1": "https://download.pytorch.org/models/resnet34-b627a593.pth",
    "resnet50_IMAGENET1K_V1": "https://download.pytorch.org/models/resnet50-0676ba61.pth",
    "resnet50_IMAGENET1K_V2": "https://download.pytorch.org/models/resnet50-11ad3fa6.pth",
    "resnet101_IMAGENET1K_V1": "https://download.pytorch.org/models/resnet101-63fe2227.pth",
    "resnet101_IMAGENET1K_V2": "https://download.pytorch.org/models/resnet101-cd907fc2.pth",
    "resnet152_IMAGENET1K_V1": "https://download.pytorch.org/models/resnet152-394f9c45.pth",
    "resnet152_IMAGENET1K_V2": "https://download.pytorch.org/models/resnet152-f82ba261.pth",
    "resnext50_32X4D_IMAGENET1K_V1": "https://download.pytorch.org/models/resnext50_32x4d-7cdf4587.pth",
    "resnext50_32X4D_IMAGENET1K_V2": "https://download.pytorch.org/models/resnext50_32x4d-1a0047aa.pth",
    "resnext101_32X8D_IMAGENET1K_V1": "https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth",
    "resnext101_32X8D_IMAGENET1K_V2": "https://download.pytorch.org/models/resnext101_32x8d-110c445d.pth",
    "resnext101_64X4D_IMAGENET1K_V1": "https://download.pytorch.org/models/resnext101_64x4d-173b62eb.pth",
    "wide_resnet50_2_IMAGENET1K_V1": "https://download.pytorch.org/models/wide_resnet50_2-95faca4d.pth",
    "wide_resnet50_2_IMAGENET1K_V2": "https://download.pytorch.org/models/wide_resnet50_2-9ba9bcbe.pth",
    "wide_resnet101_2_IMAGENET1K_V1": "https://download.pytorch.org/models/wide_resnet101_2-32ee1156.pth",
    "wide_resnet101_2_IMAGENET1K_V2": "https://download.pytorch.org/models/wide_resnet101_2-d733dc28.pth",
}


class Downsample(eqx.Module):
    conv: eqx.nn.Conv2d
    bn: BatchNorm

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        key: PRNGKeyArray,
        dtype: Any,
        axis_name: str,
    ):
        _, subkey = jax.random.split(key)
        self.conv = eqx.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=1,
            stride=stride,
            use_bias=False,
            key=subkey,
            dtype=dtype,
        )

        self.bn = BatchNorm(out_channels, axis_name=axis_name, dtype=dtype)

    def __call__(
        self,
        x: Float[Array, "c_in h w"],
        state: eqx.nn.State,
    ) -> tuple[Float[Array, "c_out*e h/s w/s"], eqx.nn.State]:
        x = self.conv(x)
        x, state = self.bn(x, state)

        return x, state


class BasicBlock(eqx.Module):
    conv1: eqx.nn.Conv2d
    bn1: BatchNorm

    conv2: eqx.nn.Conv2d
    bn2: BatchNorm

    downsample: Downsample | None
    expansion: int = eqx.field(static=True, default=1)

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        downsample: Downsample | None,
        groups: int,
        base_width: int,
        dilation: int,
        key: PRNGKeyArray,
        dtype: Any,
        axis_name: str,
    ):
        key, *subkeys = jax.random.split(key, 3)

        self.conv1 = eqx.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            use_bias=False,
            key=subkeys[0],
            dtype=dtype,
        )
        self.bn1 = BatchNorm(size=out_channels, axis_name=axis_name, dtype=dtype)

        self.conv2 = eqx.nn.Conv2d(
            out_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            use_bias=False,
            key=subkeys[1],
            dtype=dtype,
        )
        self.bn2 = BatchNorm(size=out_channels, axis_name=axis_name, dtype=dtype)

        self.downsample = downsample

    def __call__(
        self,
        x: Float[Array, "c h w"],
        state: eqx.nn.State,
        *,
        inference: bool = False,
    ):
        i = x

        x = self.conv1(x)
        x, state = self.bn1(x, state)

        x = jax.nn.relu(x)

        x = self.conv2(x)
        x, state = self.bn2(x, state)

        if self.downsample:
            i, state = self.downsample(i, state)

        x += i
        x = jax.nn.relu(x)

        return x, state


class Bottleneck(eqx.Module):
    conv1: eqx.nn.Conv2d
    bn1: BatchNorm

    conv2: eqx.nn.Conv2d
    bn2: BatchNorm

    conv3: eqx.nn.Conv2d
    bn3: BatchNorm

    downsample: Downsample | None

    expansion: int = eqx.field(static=True, default=4)

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        downsample: Downsample | None,
        groups: int,
        base_width: int,
        dilation: int,
        key: PRNGKeyArray,
        dtype: Any,
        axis_name: str,
    ):
        _, *subkeys = jax.random.split(key, 4)

        width = int(out_channels * (base_width / 64.0)) * groups
        self.conv1 = eqx.nn.Conv2d(
            in_channels,
            width,
            kernel_size=1,
            use_bias=False,
            key=subkeys[0],
            dtype=dtype,
        )
        self.bn1 = BatchNorm(width, axis_name=axis_name, dtype=dtype)

        self.conv2 = eqx.nn.Conv2d(
            width,
            width,
            kernel_size=3,
            stride=stride,
            groups=groups,
            dilation=dilation,
            padding=dilation,
            use_bias=False,
            key=subkeys[1],
            dtype=dtype,
        )

        self.bn2 = BatchNorm(width, axis_name=axis_name, dtype=dtype)

        self.conv3 = eqx.nn.Conv2d(
            width,
            out_channels * self.expansion,
            kernel_size=1,
            key=subkeys[2],
            use_bias=False,
            dtype=dtype,
        )

        self.bn3 = BatchNorm(
            out_channels * self.expansion, axis_name=axis_name, dtype=dtype
        )

        self.downsample = downsample

    def __call__(
        self,
        x: Float[Array, "c_in h w"],
        state: eqx.nn.State,
        *,
        inference: bool = False,
    ) -> tuple[Float[Array, "c_out*e h/s w/s"], eqx.nn.State]:
        i = x

        x = self.conv1(x)
        x, state = self.bn1(x, state)
        x = jax.nn.relu(x)

        x = self.conv2(x)
        x, state = self.bn2(x, state)
        x = jax.nn.relu(x)

        x = self.conv3(x)
        x, state = self.bn3(x, state)

        if self.downsample:
            i, state = self.downsample(i, state)

        x += i
        x = jax.nn.relu(x)
        return x, state


class ResNetBlock(eqx.Module):
    blocks: list[BasicBlock | Bottleneck]

    def __call__(self, x, state, *, inference: bool = False):
        for block in self.blocks:
            x, state = block(x, state, inference=inference)
        return x, state


class ResNet(eqx.Module):
    conv1: eqx.nn.Conv2d
    bn: BatchNorm
    mp: eqx.nn.MaxPool2d

    layer1: ResNetBlock
    layer2: ResNetBlock
    layer3: ResNetBlock
    layer4: ResNetBlock

    avg: eqx.nn.AdaptiveAvgPool2d
    fc: eqx.nn.Linear

    running_internal_channels: int = eqx.field(static=True, default=64)
    dilation: int = eqx.field(static=True, default=1)

    def __init__(
        self,
        block: Type[BasicBlock | Bottleneck],
        layers: list[int],
        n_classes: int,
        zero_init_residual: bool,
        groups: int,
        width_per_group: int,
        replace_stride_with_dilation: list[bool] | None,
        key: PRNGKeyArray,
        dtype: Any | None = None,
        input_channels: int = 3,
        axis_name: str = "batch",
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        key, *subkeys = jax.random.split(key, 10)

        if replace_stride_with_dilation is None:
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                f"`replace_stride_with_dilation` should either be `None` "
                f"or have a length of 3, got {replace_stride_with_dilation} instead."
            )

        self.conv1 = eqx.nn.Conv2d(
            in_channels=input_channels,
            out_channels=self.running_internal_channels,
            kernel_size=7,
            stride=2,
            padding=3,
            use_bias=False,
            key=subkeys[0],
            dtype=dtype,
        )

        self.bn = BatchNorm(
            self.running_internal_channels, axis_name=axis_name, dtype=dtype
        )
        self.mp = eqx.nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layer1 = ResNetBlock(
            self._make_layer(
                block,
                64,
                layers[0],
                stride=1,
                dilate=False,
                groups=groups,
                base_width=width_per_group,
                key=subkeys[1],
                dtype=dtype,
                axis_name=axis_name,
            )
        )
        self.layer2 = ResNetBlock(
            self._make_layer(
                block,
                128,
                layers[1],
                stride=2,
                dilate=replace_stride_with_dilation[0],
                groups=groups,
                base_width=width_per_group,
                key=subkeys[2],
                dtype=dtype,
                axis_name=axis_name,
            )
        )
        self.layer3 = ResNetBlock(
            self._make_layer(
                block,
                256,
                layers[2],
                stride=2,
                dilate=replace_stride_with_dilation[1],
                groups=groups,
                base_width=width_per_group,
                key=subkeys[3],
                dtype=dtype,
                axis_name=axis_name,
            )
        )
        self.layer4 = ResNetBlock(
            self._make_layer(
                block,
                512,
                layers[3],
                stride=2,
                dilate=replace_stride_with_dilation[2],
                groups=groups,
                base_width=width_per_group,
                key=subkeys[4],
                dtype=dtype,
                axis_name=axis_name,
            )
        )

        self.avg = eqx.nn.AdaptiveAvgPool2d(target_shape=(1, 1))
        self.fc = eqx.nn.Linear(
            512 * block.expansion, n_classes, key=subkeys[-1], dtype=dtype
        )

        if zero_init_residual:
            # todo: init last bn layer with zero weights
            pass

    def _make_layer(
        self,
        block: Type[BasicBlock | Bottleneck],
        out_channels: int,
        blocks: int,
        stride: int,
        dilate: bool,
        groups: int,
        base_width: int,
        key: PRNGKeyArray,
        dtype: Any,
        axis_name: str,
    ) -> list[BasicBlock | Bottleneck]:
        downsample = None
        previous_dilation = self.dilation

        if dilate:
            self.dilation *= stride
            stride = 1

        if (
            stride != 1
            or self.running_internal_channels != out_channels * block.expansion
        ):
            key, subkey = jax.random.split(key)
            downsample = Downsample(
                self.running_internal_channels,
                out_channels * block.expansion,
                stride,
                subkey,
                dtype=dtype,
                axis_name=axis_name,
            )
        layers = []

        key, subkey = jax.random.split(key)
        layers.append(
            block(
                in_channels=self.running_internal_channels,
                out_channels=out_channels,
                stride=stride,
                downsample=downsample,
                groups=groups,
                base_width=base_width,
                dilation=previous_dilation,
                key=subkey,
                dtype=dtype,
                axis_name=axis_name,
            )
        )

        self.running_internal_channels = out_channels * block.expansion

        for _ in range(1, blocks):
            key, subkey = jax.random.split(key)
            layers.append(
                block(
                    in_channels=self.running_internal_channels,
                    out_channels=out_channels,
                    groups=groups,
                    base_width=base_width,
                    dilation=self.dilation,
                    stride=1,
                    downsample=None,
                    key=subkey,
                    dtype=dtype,
                    axis_name=axis_name,
                )
            )

        return layers

    def __call__(
        self,
        x: Float[Array, "c h w"],
        state: eqx.nn.State | None,
        *,
        inference: bool = False,
    ) -> tuple[Float[Array, " n_classes"], eqx.nn.State]:
        assert state is not None
        x = self.conv1(x)
        x, state = self.bn(x, state)
        x = jax.nn.relu(x)
        x = self.mp(x)

        x, state = self.layer1(x, state, inference=inference)
        x, state = self.layer2(x, state, inference=inference)
        x, state = self.layer3(x, state, inference=inference)
        x, state = self.layer4(x, state, inference=inference)

        x = self.avg(x)
        x = jnp.ravel(x)

        x = self.fc(x)

        return x, state


def _with_weights(
    pytree,
    weights_name: str,
    cache: bool,
    dtype: Any,
):
    resnet, state = pytree
    dtype_str = dtype_to_str(dtype)

    weights_url = _MODELS.get(weights_name)
    if weights_url is None:
        raise ValueError(f"No weights found for {weights_name}")

    jaxonmodels_dir = os.path.expanduser("~/.jaxonmodels/models")
    os.makedirs(jaxonmodels_dir, exist_ok=True)

    cache_filename = f"{weights_name}-{dtype_str}.eqx"
    cache_filepath = str(Path(jaxonmodels_dir) / cache_filename)

    if cache:
        if os.path.exists(cache_filepath):
            return eqx.tree_deserialise_leaves(
                cache_filepath,
                (resnet, state),
            )

    weights_dir = os.path.expanduser("~/.jaxonmodels/pytorch_weights")
    os.makedirs(weights_dir, exist_ok=True)
    filename = weights_url.split("/")[-1]
    weights_file = os.path.join(weights_dir, filename)
    if not os.path.exists(weights_file):
        print(f"Downloading weights from {weights_url} to {weights_file}")
        urlretrieve(weights_url, weights_file)

    import torch

    weights_dict = torch.load(weights_file, map_location=torch.device("cpu"))
    # Handle nested state_dicts if present
    if isinstance(weights_dict, dict) and "state_dict" in weights_dict:
        weights_dict = weights_dict["state_dict"]
    elif not isinstance(weights_dict, dict):
        raise TypeError(f"Loaded weights are not a dictionary: {type(weights_dict)}")

    torchfields = state_dict_to_fields(weights_dict)
    torchfields = move_running_fields_to_the_end(torchfields)
    jaxfields, state_indices = pytree_to_fields((resnet, state))

    resnet, state = convert(
        weights_dict,
        (resnet, state),
        jaxfields,
        state_indices,
        torchfields,
        dtype=dtype,
    )

    if cache:
        eqx.tree_serialise_leaves(cache_filepath, (resnet, state))

    return resnet, state


def _resnet18(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        BasicBlock,
        [2, 2, 2, 2],
        n_classes,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _resnet34(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        BasicBlock,
        [3, 4, 6, 3],
        n_classes,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _resnet50(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        Bottleneck,
        [3, 4, 6, 3],
        n_classes,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _resnet101(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        Bottleneck,
        [3, 4, 23, 3],
        n_classes,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _resnet152(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        Bottleneck,
        [3, 8, 36, 3],
        n_classes,
        zero_init_residual=False,
        groups=1,
        width_per_group=64,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _resnext50_32x4d(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        Bottleneck,
        [3, 4, 6, 3],
        n_classes,
        zero_init_residual=False,
        groups=32,
        width_per_group=4,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _resnext101_32x8d(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        Bottleneck,
        [3, 4, 23, 3],
        n_classes,
        zero_init_residual=False,
        groups=32,
        width_per_group=8,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _resnext101_64x4d(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        Bottleneck,
        [3, 4, 23, 3],
        n_classes,
        zero_init_residual=False,
        groups=64,
        width_per_group=4,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _wide_resnet50_2(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        Bottleneck,
        [3, 4, 6, 3],
        n_classes,
        zero_init_residual=False,
        groups=1,
        width_per_group=64 * 2,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _wide_resnet101_2(key, n_classes=1000, dtype: Any = None):
    key, init_key = jax.random.split(key)
    resnet, state = eqx.nn.make_with_state(ResNet)(
        Bottleneck,
        [3, 4, 23, 3],
        n_classes,
        zero_init_residual=False,
        groups=1,
        width_per_group=64 * 2,
        replace_stride_with_dilation=None,
        key=key,
        dtype=dtype,
    )
    resnet, state = kaiming_init_conv2d(resnet, state, init_key)
    return resnet, state


def _assert_model_and_weights_fit(
    model: Literal[
        "resnet18",
        "resnet34",
        "resnet50",
        "resnet101",
        "resnet152",
        "resnext50_32x4d",
        "resnext101_32x8d",
        "resnext101_64x4d",
        "wide_resnet50_2",
        "wide_resnet101_2",
    ],
    weights: Literal[
        "resnet18_IMAGENET1K_V1",
        "resnet34_IMAGENET1K_V1",
        "resnet50_IMAGENET1K_V1",
        "resnet50_IMAGENET1K_V2",
        "resnet101_IMAGENET1K_V1",
        "resnet101_IMAGENET1K_V2",
        "resnet152_IMAGENET1K_V1",
        "resnet152_IMAGENET1K_V2",
        "resnext50_32X4D_IMAGENET1K_V1",
        "resnext50_32X4D_IMAGENET1K_V2",
        "resnext101_32X8D_IMAGENET1K_V1",
        "resnext101_32X8D_IMAGENET1K_V2",
        "resnext101_64X4D_IMAGENET1K_V1",
        "wide_resnet50_2_IMAGENET1K_V1",
        "wide_resnet50_2_IMAGENET1K_V2",
        "wide_resnet101_2_IMAGENET1K_V1",
        "wide_resnet101_2_IMAGENET1K_V2",
    ],
):
    # Improved check: Handles ResNeXt and Wide ResNet naming
    weights_prefix = weights.split("_IMAGENET")[0].lower().replace("_", "")
    model_prefix = model.lower().replace("_", "")

    # Specific check for 'x' vs 'X' in resnext
    if "resnext" in model:
        weights_prefix = weights_prefix.replace("x", "")
        model_prefix = model_prefix.replace("x", "")

    if not model_prefix.startswith(weights_prefix):
        raise ValueError(
            f"Model '{model}' is incompatible with weights '{weights}'. "
            f"Weight prefix '{weights_prefix}' "
            f"does not match model prefix '{model_prefix}'."
        )


def load_resnet(
    model: Literal[
        "resnet18",
        "resnet34",
        "resnet50",
        "resnet101",
        "resnet152",
        "resnext50_32x4d",
        "resnext101_32x8d",
        "resnext101_64x4d",
        "wide_resnet50_2",
        "wide_resnet101_2",
    ],
    weights: Literal[
        "resnet18_IMAGENET1K_V1",
        "resnet34_IMAGENET1K_V1",
        "resnet50_IMAGENET1K_V1",
        "resnet50_IMAGENET1K_V2",
        "resnet101_IMAGENET1K_V1",
        "resnet101_IMAGENET1K_V2",
        "resnet152_IMAGENET1K_V1",
        "resnet152_IMAGENET1K_V2",
        "resnext50_32X4D_IMAGENET1K_V1",
        "resnext50_32X4D_IMAGENET1K_V2",
        "resnext101_32X8D_IMAGENET1K_V1",
        "resnext101_32X8D_IMAGENET1K_V2",
        "resnext101_64X4D_IMAGENET1K_V1",
        "wide_resnet50_2_IMAGENET1K_V1",
        "wide_resnet50_2_IMAGENET1K_V2",
        "wide_resnet101_2_IMAGENET1K_V1",
        "wide_resnet101_2_IMAGENET1K_V2",
    ]
    | None = None,
    n_classes: int = 1000,
    cache: bool = True,
    *,
    key: PRNGKeyArray | None = None,
    dtype: Any | None = None,  # Added optional dtype parameter
) -> tuple[ResNet, eqx.nn.State]:
    """
    Load a ResNet model with optional pre-trained weights.

    Args:
        model: The name of the ResNet model architecture to load.
        weights: The specific pre-trained weights to load
                (e.g., "resnet50_IMAGENET1K_V2").
                 If None, the model is initialized randomly (with Kaiming init).
        n_classes: Number of output classes for the final fully connected layer.
        cache: If True, attempts to load/save the JAX model with weights from/to
        a local cache
               (~/.jaxonmodels/models). Cache filenames include the dtype.
        key: JAX random key for initialization. Defaults to jax.random.key(42).
        dtype: The data type for the model's parameters
        (e.g., jnp.float32, jnp.bfloat16).
               Defaults to `jaxonmodels.functions.default_floating_dtype()`.

    Returns:
        A tuple containing the ResNet model (eqx.Module)
        and its initial state (eqx.nn.State).
    """
    if key is None:
        key = jax.random.key(42)

    # Determine default dtype if not provided
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None

    resnet, state = None, None

    match model:
        case "resnet18":
            resnet, state = _resnet18(key, n_classes, dtype=dtype)
        case "resnet34":
            resnet, state = _resnet34(key, n_classes, dtype=dtype)
        case "resnet50":
            resnet, state = _resnet50(key, n_classes, dtype=dtype)
        case "resnet101":
            resnet, state = _resnet101(key, n_classes, dtype=dtype)
        case "resnet152":
            resnet, state = _resnet152(key, n_classes, dtype=dtype)
        case "resnext50_32x4d":
            resnet, state = _resnext50_32x4d(key, n_classes, dtype=dtype)
        case "resnext101_32x8d":
            resnet, state = _resnext101_32x8d(key, n_classes, dtype=dtype)
        case "resnext101_64x4d":
            resnet, state = _resnext101_64x4d(key, n_classes, dtype=dtype)
        case "wide_resnet50_2":
            resnet, state = _wide_resnet50_2(key, n_classes, dtype=dtype)
        case "wide_resnet101_2":
            resnet, state = _wide_resnet101_2(key, n_classes, dtype=dtype)
        case _:
            raise ValueError(f"Unknown model name: {model}")

    if weights is not None:
        _assert_model_and_weights_fit(model, weights)
        resnet, state = _with_weights((resnet, state), weights, cache, dtype=dtype)

    assert resnet is not None
    assert state is not None
    return resnet, state



import functools
import os
from pathlib import Path
from urllib.request import urlretrieve

import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any, Callable, Literal
from equinox.nn import StatefulLayer
from jaxonlayers.functions import shifted_window_attention
from jaxonlayers.layers import (
    LayerNorm,
    StochasticDepth,
)
from jaxonlayers.layers.abstract import AbstractNorm, AbstractNormStateful
from jaxonlayers.layers.sequential import BatchedLinear
from jaxtyping import Array, Float, PRNGKeyArray, PyTree
from statedict2pytree import (
    convert,
    pytree_to_fields,
    state_dict_to_fields,
)

from jaxonmodels.functions import patch_merging_pad
from jaxonmodels.functions.utils import default_floating_dtype, dtype_to_str
from jaxonmodels.statedict2pytree.model_orders import get_swin_model_order

_SWIN_MODELS = {
    "swin_t_IMAGENET1K_V1": "https://download.pytorch.org/models/swin_t-704ceda3.pth",
    "swin_s_IMAGENET1K_V1": "https://download.pytorch.org/models/swin_s-5e29d889.pth",
    "swin_b_IMAGENET1K_V1": "https://download.pytorch.org/models/swin_b-68c6b09e.pth",
    "swin_v2_t_IMAGENET1K_V1": "https://download.pytorch.org/models/swin_v2_t-b137f0e2.pth",
    "swin_v2_s_IMAGENET1K_V1": "https://download.pytorch.org/models/swin_v2_s-637d8ceb.pth",
    "swin_v2_b_IMAGENET1K_V1": "https://download.pytorch.org/models/swin_v2_b-781e5279.pth",
}


def get_relative_position_bias(
    relative_position_bias_table: Array,
    relative_position_index: Array,
    window_size: list[int],
) -> Array:
    N = window_size[0] * window_size[1]
    relative_position_bias = relative_position_bias_table[relative_position_index]
    relative_position_bias = relative_position_bias.reshape(N, N, -1)
    relative_position_bias = jnp.transpose(relative_position_bias, (2, 0, 1))
    return relative_position_bias


class MLPWithDropout(eqx.nn.StatefulLayer):
    layers: list[PyTree]

    def __init__(
        self,
        in_channels: int,
        hidden_channels: list[int],
        key: PRNGKeyArray,
        norm_layer: Callable[..., AbstractNorm] | None,
        activation: Callable | None,
        use_bias: bool,
        dropout: float,
        inference: bool,
        dtype: Any,
    ):
        assert dtype is not None

        key, *subkeys = jax.random.split(key, len(hidden_channels) + 1)
        layers = []
        in_dim = in_channels
        for i, hidden_dim in enumerate(hidden_channels[:-1]):
            layers.append(
                BatchedLinear(
                    in_dim, hidden_dim, use_bias=use_bias, key=subkeys[i], dtype=dtype
                )
            )
            if norm_layer is not None:
                layers.append(norm_layer(hidden_dim))
            layers.append(
                eqx.nn.Lambda(fn=activation if activation is not None else lambda x: x)
            )
            layers.append(eqx.nn.Dropout(dropout, inference=inference))
            in_dim = hidden_dim

        layers.append(
            BatchedLinear(
                in_dim,
                hidden_channels[-1],
                use_bias=use_bias,
                dtype=dtype,
                key=subkeys[-1],
            )
        )
        layers.append(eqx.nn.Dropout(dropout, inference=inference))
        self.layers = layers

    def __call__(
        self, x: Float[Array, "H W C"], state: eqx.nn.State
    ) -> tuple[Array, eqx.nn.State]:
        for layer in self.layers:
            if isinstance(layer, eqx.nn.StatefulLayer):
                x, state = layer(x, state)
            else:
                x = layer(x)

        return x, state


class PatchMerging(StatefulLayer):
    reduction: BatchedLinear
    norm: AbstractNorm | AbstractNormStateful

    inference: bool

    def __init__(
        self,
        dim: int,
        norm_layer: Callable[..., AbstractNorm | AbstractNormStateful],
        inference: bool,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        self.inference = inference
        self.reduction = BatchedLinear(
            4 * dim, 2 * dim, use_bias=False, key=key, dtype=dtype
        )
        self.norm = norm_layer(4 * dim)

    def __call__(
        self,
        x: Float[Array, "H W C"],
        state: eqx.nn.State,
        *,
        key: PRNGKeyArray | None = None,
    ) -> tuple[Float[Array, "H_half W_half C*2"], eqx.nn.State]:
        x = patch_merging_pad(x)
        if isinstance(self.norm, StatefulLayer):
            x, state = self.norm(x, state=state)
        else:
            x = self.norm(x)
        x = self.reduction(x)  # ... H/2 W/2 2*C
        return x, state


class PatchMergingV2(StatefulLayer):
    reduction: BatchedLinear
    norm: AbstractNorm | AbstractNormStateful

    inference: bool

    def __init__(
        self,
        dim: int,
        norm_layer: Callable[..., AbstractNorm | AbstractNormStateful],
        inference: bool,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        self.inference = inference
        self.reduction = BatchedLinear(
            4 * dim, 2 * dim, use_bias=False, key=key, dtype=dtype
        )
        self.norm = norm_layer(2 * dim)

    def __call__(
        self,
        x: Float[Array, "H W C"],
        state: eqx.nn.State,
        *,
        key: PRNGKeyArray | None = None,
    ) -> tuple[Float[Array, "H_half W_half C*2"], eqx.nn.State]:
        x = patch_merging_pad(x)
        x = self.reduction(x)  # ... H/2 W/2 2*C
        if isinstance(self.norm, StatefulLayer):
            x, state = self.norm(x, state=state)
        else:
            x = self.norm(x)
        return x, state


class ShiftedWindowAttention(eqx.nn.StatefulLayer):
    logit_scale: Array | None
    relative_position_bias_table: Array | None
    relative_position_index: eqx.nn.StateIndex

    relative_coords_table: eqx.nn.StateIndex | None

    qkv: eqx.nn.Linear
    proj: eqx.nn.Linear

    cpb_mlp_ln1: BatchedLinear | None
    cpb_mlp_ln2: BatchedLinear | None
    inference: bool

    window_size: list[int] = eqx.field(static=True)
    shift_size: list[int] = eqx.field(static=True)
    num_heads: int = eqx.field(static=True)
    attention_dropout: float = eqx.field(static=True)
    dropout: float = eqx.field(static=True)

    def __init__(
        self,
        dim: int,
        window_size: list[int],
        shift_size: list[int],
        num_heads: int,
        qkv_bias: bool,
        proj_bias: bool,
        attention_dropout: float,
        dropout: float,
        key: PRNGKeyArray,
        inference: bool,
        dtype: Any,
    ):
        self.inference = inference
        if len(window_size) != 2 or len(shift_size) != 2:
            raise ValueError("window_size and shift_size must be of length 2")
        self.window_size = window_size
        self.shift_size = shift_size
        self.num_heads = num_heads
        self.attention_dropout = attention_dropout
        self.dropout = dropout

        key, subkey1, subkey2 = jax.random.split(key, 3)
        self.qkv = eqx.nn.Linear(
            dim, dim * 3, use_bias=qkv_bias, key=subkey1, dtype=dtype
        )
        self.proj = eqx.nn.Linear(
            dim, dim, use_bias=proj_bias, key=subkey2, dtype=dtype
        )

        self.relative_position_bias_table = jnp.zeros(
            shape=(
                (2 * self.window_size[0] - 1) * (2 * self.window_size[1] - 1),
                self.num_heads,
            ),
            dtype=dtype,
        )

        # get pair-wise relative position index for each token inside the window
        coords_h = jnp.arange(self.window_size[0])
        coords_w = jnp.arange(self.window_size[1])
        coords = jnp.stack(jnp.meshgrid(coords_h, coords_w, indexing="ij"))  # 2, Wh, Ww
        coords_flatten = jnp.reshape(coords, (coords.shape[0], -1))  # 2, Wh*Ww
        relative_coords = (
            coords_flatten[:, :, None] - coords_flatten[:, None, :]
        )  # 2, Wh*Ww, Wh*Ww
        relative_coords = jnp.transpose(relative_coords, (1, 2, 0))  # Wh*Ww, Wh*Ww, 2
        relative_coords = relative_coords.at[:, :, 0].set(
            relative_coords[:, :, 0] + self.window_size[0] - 1
        )  # shift to start from 0
        relative_coords = relative_coords.at[:, :, 1].set(
            relative_coords[:, :, 1] + self.window_size[1] - 1
        )
        relative_coords = relative_coords.at[:, :, 0].set(
            relative_coords[:, :, 0] * 2 * self.window_size[1] - 1
        )
        relative_position_index = relative_coords.sum(-1).flatten()  # Wh*Ww*Wh*Ww
        self.relative_position_index = eqx.nn.StateIndex(relative_position_index)

        self.cpb_mlp_ln1 = None
        self.cpb_mlp_ln2 = None
        self.logit_scale = None
        self.relative_coords_table = None

    def get_relative_position_bias(self, state: eqx.nn.State) -> Array:
        assert self.relative_position_bias_table is not None
        return get_relative_position_bias(
            self.relative_position_bias_table,
            state.get(self.relative_position_index),
            self.window_size,
        )

    def __call__(
        self, x: Float[Array, "H W C"], state: eqx.nn.State
    ) -> tuple[Float[Array, "H W C"], eqx.nn.State]:
        relative_position_bias = self.get_relative_position_bias(state)
        return shifted_window_attention(
            x,
            self.qkv.weight,
            self.proj.weight,
            relative_position_bias,
            self.window_size,
            self.num_heads,
            shift_size=self.shift_size,
            attention_dropout=self.attention_dropout,
            dropout=self.dropout,
            qkv_bias=self.qkv.bias,
            proj_bias=self.proj.bias,
            inference=self.inference,
        ), state


# TODO: This is not the Eqx way of doing things!
class ShiftedWindowAttentionV2(ShiftedWindowAttention):
    def __init__(
        self,
        dim: int,
        window_size: list[int],
        shift_size: list[int],
        num_heads: int,
        qkv_bias: bool,
        proj_bias: bool,
        attention_dropout: float,
        dropout: float,
        key: PRNGKeyArray,
        inference: bool,
        dtype: Any,
    ):
        super().__init__(
            dim,
            window_size,
            shift_size,
            num_heads,
            qkv_bias=qkv_bias,
            proj_bias=proj_bias,
            attention_dropout=attention_dropout,
            dropout=dropout,
            key=key,
            inference=inference,
            dtype=dtype,
        )

        self.logit_scale = jnp.log(10 * jnp.ones((num_heads, 1, 1)))
        # mlp to generate continuous relative position bias
        key, subkey1, subkey2 = jax.random.split(key, 3)
        self.cpb_mlp_ln1 = BatchedLinear(
            2, 512, use_bias=True, key=subkey1, dtype=dtype
        )

        self.cpb_mlp_ln2 = BatchedLinear(
            512, num_heads, use_bias=False, key=subkey2, dtype=dtype
        )
        if qkv_bias:
            assert self.qkv.bias is not None
            length = self.qkv.bias.size // 3
            new_bias = self.qkv.bias.at[length : 2 * length].set(
                jnp.zeros_like(self.qkv.bias[length : 2 * length])
            )
            where = lambda x: x.bias
            self.qkv = eqx.tree_at(where, self.qkv, new_bias)

        # get relative_coords_table
        relative_coords_h = jnp.arange(
            -(self.window_size[0] - 1), self.window_size[0], dtype=dtype
        )
        relative_coords_w = jnp.arange(
            -(self.window_size[1] - 1), self.window_size[1], dtype=dtype
        )
        relative_coords_table = jnp.stack(
            jnp.meshgrid(relative_coords_h, relative_coords_w, indexing="ij")
        )
        relative_coords_table = jnp.expand_dims(
            jnp.transpose(relative_coords_table, (1, 2, 0)), axis=0
        )  # 1, 2*Wh-1, 2*Ww-1, 2

        relative_coords_table = relative_coords_table.at[:, :, :, 0].set(
            relative_coords_table[:, :, :, 0] / self.window_size[0] - 1
        )
        relative_coords_table = relative_coords_table.at[:, :, :, 1].set(
            relative_coords_table[:, :, :, 1] / self.window_size[1] - 1
        )

        relative_coords_table *= 8  # normalize to -8, 8
        relative_coords_table = (
            jnp.sign(relative_coords_table)
            * jnp.log2(jnp.abs(relative_coords_table) + 1.0)
            / 3.0
        )

        self.relative_coords_table = eqx.nn.StateIndex(relative_coords_table)
        self.relative_position_bias_table = None

    def get_relative_position_bias(self, state: eqx.nn.State) -> Array:
        assert self.relative_coords_table is not None
        assert self.cpb_mlp_ln1 is not None
        assert self.cpb_mlp_ln2 is not None
        relative_coords_table = state.get(self.relative_coords_table)
        relative_position_index = state.get(self.relative_position_index)

        cpb = self.cpb_mlp_ln1(relative_coords_table)
        cpb = jax.nn.relu(cpb)
        cpb = self.cpb_mlp_ln2(cpb)

        relative_position_bias = get_relative_position_bias(
            cpb.reshape(-1, self.num_heads),
            relative_position_index,
            self.window_size,
        )
        relative_position_bias = 16 * jax.nn.sigmoid(relative_position_bias)
        return relative_position_bias

    def __call__(
        self, x: Float[Array, "H W C"], state: eqx.nn.State
    ) -> tuple[Float[Array, "H W C"], eqx.nn.State]:
        relative_position_bias = self.get_relative_position_bias(state)
        return shifted_window_attention(
            x,
            self.qkv.weight,
            self.proj.weight,
            relative_position_bias,
            self.window_size,
            self.num_heads,
            shift_size=self.shift_size,
            attention_dropout=self.attention_dropout,
            dropout=self.dropout,
            qkv_bias=self.qkv.bias,
            proj_bias=self.proj.bias,
            logit_scale=self.logit_scale,
            inference=self.inference,
        ), state


class SwinTransformerBlock(StatefulLayer):
    norm1: AbstractNorm
    attn: eqx.Module

    stochastic_depth: StochasticDepth
    norm2: AbstractNorm
    mlp: MLPWithDropout

    def __init__(
        self,
        dim: int,
        num_heads: int,
        window_size: list[int],
        shift_size: list[int],
        mlp_ratio: float,
        dropout: float,
        attention_dropout: float,
        stochastic_depth_prob: float,
        norm_layer: Callable[..., AbstractNorm],
        attn_layer: Callable[..., eqx.Module],
        key: PRNGKeyArray,
        inference: bool,
        dtype: Any,
    ):
        self.norm1 = norm_layer(dim)
        key, subkey = jax.random.split(key)
        self.attn = attn_layer(
            dim,
            window_size,
            shift_size,
            num_heads,
            attention_dropout=attention_dropout,
            dropout=dropout,
            qkv_bias=True,
            proj_bias=True,
            key=subkey,
            inference=inference,
            dtype=dtype,
        )
        self.stochastic_depth = StochasticDepth(
            stochastic_depth_prob, "row", inference=inference
        )
        self.norm2 = norm_layer(dim)
        self.mlp = MLPWithDropout(
            dim,
            [int(dim * mlp_ratio), dim],
            norm_layer=None,
            use_bias=True,
            activation=jax.nn.gelu,
            dropout=dropout,
            inference=inference,
            key=key,
            dtype=dtype,
        )

        # todo: init properly

    def __call__(
        self, x: Float[Array, "H W C"], state: eqx.nn.State, key: PRNGKeyArray
    ) -> tuple[Array, eqx.nn.State]:
        key, subkey = jax.random.split(key)
        x_norm = eqx.filter_vmap(eqx.filter_vmap(self.norm1))(x)
        attn, state = self.attn(x_norm, state)  # pyright: ignore # ty:ignore
        x = x + self.stochastic_depth(attn, key=key)
        x_norm = eqx.filter_vmap(eqx.filter_vmap(self.norm2))(x)
        mlp_o, state = self.mlp(x_norm, state)
        x = x + self.stochastic_depth(mlp_o, key=subkey)
        return x, state


class SwinTransformerBlockV2(SwinTransformerBlock):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        window_size: list[int],
        shift_size: list[int],
        mlp_ratio: float,
        dropout: float,
        attention_dropout: float,
        stochastic_depth_prob: float,
        norm_layer: Callable[..., AbstractNorm],
        attn_layer: Callable[..., eqx.Module],
        key: PRNGKeyArray,
        inference: bool,
        dtype: Any,
    ):
        super().__init__(
            dim=dim,
            num_heads=num_heads,
            window_size=window_size,
            shift_size=shift_size,
            mlp_ratio=mlp_ratio,
            dropout=dropout,
            attention_dropout=attention_dropout,
            stochastic_depth_prob=stochastic_depth_prob,
            norm_layer=norm_layer,
            attn_layer=attn_layer,
            key=key,
            inference=inference,
            dtype=dtype,
        )

    def __call__(
        self, x: Float[Array, "H W C"], state: eqx.nn.State, key: PRNGKeyArray
    ) -> tuple[Array, eqx.nn.State]:
        key, subkey = jax.random.split(key)

        # Apply attention first, then norm (different from V1)
        attn, state = self.attn(x, state)  # ty:ignore[call-non-callable]
        attn_norm = eqx.filter_vmap(eqx.filter_vmap(self.norm1))(attn)
        x = x + self.stochastic_depth(attn_norm, key=key)

        # Apply MLP first, then norm (different from V1)
        mlp_o, state = self.mlp(x, state)
        mlp_norm = eqx.filter_vmap(eqx.filter_vmap(self.norm2))(mlp_o)
        x = x + self.stochastic_depth(mlp_norm, key=subkey)

        return x, state


class SwinTransformer(eqx.Module):
    features: eqx.nn.Sequential
    norm: eqx.Module
    avgpool: eqx.nn.AdaptiveAvgPool2d
    head: eqx.nn.Linear

    v: int = eqx.field(static=True)

    def __init__(
        self,
        patch_size: list[int],
        embed_dim: int,
        depths: list[int],
        num_heads: list[int],
        window_size: list[int],
        mlp_ratio: float,
        dropout: float,
        attention_dropout: float,
        stochastic_depth_prob: float,
        num_classes: int,
        norm_layer: Callable[..., AbstractNorm] | None,
        block: Callable[..., eqx.Module] | None,
        downsample_layer: Callable[..., eqx.Module],
        attn_layer: Callable[..., eqx.Module],
        inference: bool,
        key: PRNGKeyArray,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        if block is None:
            block = SwinTransformerBlock
        if block is SwinTransformerBlockV2:
            self.v = 2
        else:
            self.v = 1
        if norm_layer is None:
            norm_layer = functools.partial(LayerNorm, eps=1e-5, dtype=dtype)

        layers = []
        # split image into non-overlapping patches
        key, subkey = jax.random.split(key)
        fn = functools.partial(jnp.transpose, axes=[1, 2, 0])
        layers.append(
            eqx.nn.Sequential(
                [
                    eqx.nn.Conv2d(
                        3,
                        embed_dim,
                        kernel_size=(patch_size[0], patch_size[1]),
                        stride=(patch_size[0], patch_size[1]),
                        key=subkey,
                        dtype=dtype,
                    ),
                    eqx.nn.Lambda(fn=fn),
                    norm_layer(embed_dim),  # pyright: ignore
                ],
            )
        )

        total_stage_blocks = sum(depths)
        stage_block_id = 0
        # build SwinTransformer blocks
        for i_stage in range(len(depths)):
            stage = []
            dim = embed_dim * 2**i_stage
            for i_layer in range(depths[i_stage]):
                sd_prob = (
                    stochastic_depth_prob
                    * float(stage_block_id)
                    / (total_stage_blocks - 1)
                )
                key, subkey = jax.random.split(key)
                stage.append(
                    block(
                        dim,
                        num_heads[i_stage],
                        window_size=window_size,
                        shift_size=[
                            0 if i_layer % 2 == 0 else w // 2 for w in window_size
                        ],
                        mlp_ratio=mlp_ratio,
                        dropout=dropout,
                        attention_dropout=attention_dropout,
                        stochastic_depth_prob=sd_prob,
                        norm_layer=norm_layer,
                        attn_layer=attn_layer,
                        key=key,
                        inference=inference,
                        dtype=dtype,
                    )
                )
                stage_block_id += 1
            layers.append(eqx.nn.Sequential(stage))
            # add patch merging layer
            if i_stage < (len(depths) - 1):
                key, subkey = jax.random.split(key)
                layers.append(
                    downsample_layer(
                        dim,
                        norm_layer,
                        dtype=dtype,
                        inference=inference,
                        key=subkey,
                    )
                )
        self.features = eqx.nn.Sequential(layers)

        num_features = embed_dim * 2 ** (len(depths) - 1)
        self.norm = norm_layer(num_features)
        self.avgpool = eqx.nn.AdaptiveAvgPool2d(1)
        key, subkey = jax.random.split(key)
        self.head = eqx.nn.Linear(num_features, num_classes, key=subkey, dtype=dtype)

        # todo: init properly

    def __call__(
        self, x: Array, state: eqx.nn.State, key: PRNGKeyArray
    ) -> tuple[Array, eqx.nn.State]:
        x, state = self.features(x, state, key=key)
        x = eqx.filter_vmap(eqx.filter_vmap(self.norm))(x)  # pyright: ignore
        x = jnp.transpose(x, (2, 0, 1))
        x = self.avgpool(x)
        x = jnp.ravel(x)
        x = self.head(x)
        return x, state

    def model_order(self) -> list[str]:
        return get_swin_model_order(self.v)


def _swin_transformer(
    key: PRNGKeyArray,
    inference: bool,
    num_classes: int,
    patch_size: list[int],
    embed_dim: int,
    depths: list[int],
    num_heads: list[int],
    window_size: list[int],
    stochastic_depth_prob: float,
    block: type[StatefulLayer],
    downsample_layer: type[PatchMerging | PatchMergingV2 | eqx.Module],
    attn_layer: type[ShiftedWindowAttention | ShiftedWindowAttentionV2 | eqx.Module],
    norm_layer: Callable | None = None,
    dtype: Any = None,
) -> tuple[SwinTransformer, eqx.nn.State]:
    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None

    if norm_layer is None:
        norm_layer = functools.partial(LayerNorm, eps=1e-5, dtype=dtype)

    model, state = eqx.nn.make_with_state(SwinTransformer)(
        patch_size=patch_size,
        embed_dim=embed_dim,
        depths=depths,
        num_heads=num_heads,
        window_size=window_size,
        mlp_ratio=4.0,
        dropout=0.0,
        attention_dropout=0.0,
        stochastic_depth_prob=stochastic_depth_prob,
        num_classes=num_classes,
        norm_layer=norm_layer,
        block=block,
        downsample_layer=downsample_layer,
        attn_layer=attn_layer,
        inference=inference,
        key=key,
        dtype=dtype,
    )
    return model, state


# Swin V1 Variants
def _swin_t(key, inference: bool, num_classes=1000, dtype: Any = None):
    return _swin_transformer(
        key=key,
        inference=inference,
        num_classes=num_classes,
        dtype=dtype,
        patch_size=[4, 4],
        embed_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[7, 7],
        stochastic_depth_prob=0.2,
        block=SwinTransformerBlock,
        downsample_layer=PatchMerging,
        attn_layer=ShiftedWindowAttention,
    )


def _swin_s(key, inference: bool, num_classes=1000, dtype: Any = None):
    return _swin_transformer(
        key=key,
        inference=inference,
        num_classes=num_classes,
        dtype=dtype,
        patch_size=[4, 4],
        embed_dim=96,
        depths=[2, 2, 18, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[7, 7],
        stochastic_depth_prob=0.3,
        block=SwinTransformerBlock,
        downsample_layer=PatchMerging,
        attn_layer=ShiftedWindowAttention,
    )


def _swin_b(key, inference: bool, num_classes=1000, dtype: Any = None):
    return _swin_transformer(
        key=key,
        inference=inference,
        num_classes=num_classes,
        dtype=dtype,
        patch_size=[4, 4],
        embed_dim=128,
        depths=[2, 2, 18, 2],
        num_heads=[4, 8, 16, 32],
        window_size=[7, 7],
        stochastic_depth_prob=0.5,
        block=SwinTransformerBlock,
        downsample_layer=PatchMerging,
        attn_layer=ShiftedWindowAttention,
    )


def _swin_v2_t(key, inference: bool, num_classes=1000, dtype: Any = None):
    return _swin_transformer(
        key=key,
        inference=inference,
        num_classes=num_classes,
        dtype=dtype,
        patch_size=[4, 4],
        embed_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[8, 8],
        stochastic_depth_prob=0.2,
        block=SwinTransformerBlockV2,
        downsample_layer=PatchMergingV2,
        attn_layer=ShiftedWindowAttentionV2,
    )


def _swin_v2_s(key, inference: bool, num_classes=1000, dtype: Any = None):
    return _swin_transformer(
        key=key,
        inference=inference,
        num_classes=num_classes,
        dtype=dtype,
        patch_size=[4, 4],
        embed_dim=96,
        depths=[2, 2, 18, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[8, 8],
        stochastic_depth_prob=0.3,
        block=SwinTransformerBlockV2,
        downsample_layer=PatchMergingV2,
        attn_layer=ShiftedWindowAttentionV2,
    )


def _swin_v2_b(key, inference: bool, num_classes=1000, dtype: Any = None):
    return _swin_transformer(
        key=key,
        inference=inference,
        num_classes=num_classes,
        dtype=dtype,
        patch_size=[4, 4],
        embed_dim=128,
        depths=[2, 2, 18, 2],
        num_heads=[4, 8, 16, 32],
        window_size=[8, 8],
        stochastic_depth_prob=0.5,
        block=SwinTransformerBlockV2,
        downsample_layer=PatchMergingV2,
        attn_layer=ShiftedWindowAttentionV2,
    )


def _swin_with_weights(
    model: PyTree,
    weights: str,
    cache: bool,
    dtype: Any,
) -> tuple[SwinTransformer, eqx.nn.State]:
    dtype_str = dtype_to_str(dtype)

    weights_url = _SWIN_MODELS.get(weights)
    if weights_url is None:
        raise ValueError(f"No weights found for {weights}")

    jaxonmodels_dir = os.path.expanduser("~/.jaxonmodels/models")
    os.makedirs(jaxonmodels_dir, exist_ok=True)

    cache_filename = f"{weights}-{dtype_str}.eqx"
    cache_filepath = str(Path(jaxonmodels_dir) / cache_filename)

    if cache and os.path.exists(cache_filepath):
        print(f"Loading cached JAX model from: {cache_filepath}")
        return eqx.tree_deserialise_leaves(cache_filepath, model)  # pyright: ignore

    weights_dir = os.path.expanduser("~/.jaxonmodels/pytorch_weights")
    os.makedirs(weights_dir, exist_ok=True)
    filename = weights_url.split("/")[-1]
    weights_file = os.path.join(weights_dir, filename)
    if not os.path.exists(weights_file):
        print(f"Downloading weights from {weights_url} to {weights_file}")
        urlretrieve(weights_url, weights_file)
    else:
        print(f"Using existing PyTorch weights file: {weights_file}")

    print("Loading PyTorch state dict...")
    import torch

    weights_dict = torch.load(weights_file, map_location=torch.device("cpu"))
    if isinstance(weights_dict, dict) and "model" in weights_dict:
        weights_dict = weights_dict["model"]
    elif not isinstance(weights_dict, dict):
        raise TypeError(f"Expected state_dict but got {type(weights_dict)}")

    print("Converting PyTorch weights to JAX format...")

    torchfields = state_dict_to_fields(weights_dict)
    jaxfields, state_indices = pytree_to_fields(model)

    model = convert(
        weights_dict,
        model,
        jaxfields,
        state_indices,
        torchfields,
        dtype=dtype,
    )

    if cache:
        eqx.tree_serialise_leaves(cache_filepath, model)

    return model


def _assert_model_and_weights_fit_swin(model_name: str, weights_name: str):
    if "_" in weights_name:
        weights_prefix = "_".join(weights_name.split("_")[:2])
        if weights_prefix.startswith("swin_v2"):
            weights_model = "_".join(weights_name.split("_")[:3])
        else:
            weights_model = weights_prefix

        if weights_model != model_name:
            raise ValueError(
                f"Model {model_name} is incompatible with weights {weights_name}. "
                f"Weight prefix '{weights_model}' does not match model name."
            )


def load_swin_transformer(
    model_name: Literal[
        "swin_t",
        "swin_s",
        "swin_b",  # V1
        "swin_v2_t",
        "swin_v2_s",
        "swin_v2_b",  # V2
    ],
    weights: Literal[
        "swin_t_IMAGENET1K_V1",
        "swin_s_IMAGENET1K_V1",
        "swin_b_IMAGENET1K_V1",
        "swin_v2_t_IMAGENET1K_V1",
        "swin_v2_s_IMAGENET1K_V1",
        "swin_v2_b_IMAGENET1K_V1",
    ]
    | None = None,
    num_classes: int = 1000,
    cache: bool = True,
    inference: bool = True,
    key: PRNGKeyArray | None = None,
    dtype: Any = None,
) -> tuple[SwinTransformer, eqx.nn.State]:
    if key is None:
        key = jax.random.key(42)

    if dtype is None:
        dtype = default_floating_dtype()
    assert dtype is not None

    model = None

    match model_name:
        case "swin_t":
            model = _swin_t(key, inference, num_classes, dtype=dtype)
        case "swin_s":
            model = _swin_s(key, inference, num_classes, dtype=dtype)
        case "swin_b":
            model = _swin_b(key, inference, num_classes, dtype=dtype)
        case "swin_v2_t":
            model = _swin_v2_t(key, inference, num_classes, dtype=dtype)
        case "swin_v2_s":
            model = _swin_v2_s(key, inference, num_classes, dtype=dtype)
        case "swin_v2_b":
            model = _swin_v2_b(key, inference, num_classes, dtype=dtype)
        case _:
            raise ValueError(f"Unknown Swin Transformer model name: {model_name}")

    if weights:
        _assert_model_and_weights_fit_swin(model_name, weights)
        model = _swin_with_weights(model, weights, cache, dtype=dtype)

    assert model is not None
    return model



import os
from pathlib import Path
from urllib.request import urlretrieve

import equinox as eqx
import jax
import jax.numpy as jnp
import jaxonlayers.functions as F
from beartype.typing import Any, Literal
from jaxonlayers.layers import BatchNorm, MultiheadAttention
from jaxtyping import Array, Float, Int, PRNGKeyArray, PyTree
from statedict2pytree import (
    convert,
    move_running_fields_to_the_end,
    pytree_to_fields,
    state_dict_to_fields,
)

from jaxonmodels.functions import dtype_to_str

_MODELS = {
    "RN50": "https://openaipublic.azureedge.net/clip/models/afeb0e10f9e5a86da6080e35cf09123aca3b358a0c3e3b6c78a7b63bc04b6762/RN50.pt",
    "RN101": "https://openaipublic.azureedge.net/clip/models/8fa8567bab74a42d41c5915025a8e4538c3bdbe8804a470a72f30b0d94fab599/RN101.pt",
    "RN50x4": "https://openaipublic.azureedge.net/clip/models/7e526bd135e493cef0776de27d5f42653e6b4c8bf9e0f653bb11773263205fdd/RN50x4.pt",
    "RN50x16": "https://openaipublic.azureedge.net/clip/models/52378b407f34354e150460fe41077663dd5b39c54cd0bfd2b27167a4a06ec9aa/RN50x16.pt",
    "RN50x64": "https://openaipublic.azureedge.net/clip/models/be1cfb55d75a9666199fb2206c106743da0f6468c9d327f3e0d0a543a9919d9c/RN50x64.pt",
    "ViT-B/32": "https://openaipublic.azureedge.net/clip/models/40d365715913c9da98579312b702a82c18be219cc2a73407c4526f58eba950af/ViT-B-32.pt",
    "ViT-B/16": "https://openaipublic.azureedge.net/clip/models/5806e77cd80f8b59890b7e101eabd078d9fb84e6937f9e85e4ecb61988df416f/ViT-B-16.pt",
    "ViT-L/14": "https://openaipublic.azureedge.net/clip/models/b8cca3fd41ae0c99ba7e8951adf17d267cdb84cd88be6f7c2e0eca1737a03836/ViT-L-14.pt",
    "ViT-L/14@336px": "https://openaipublic.azureedge.net/clip/models/3035c92b350959924f9f00213499208652fc7ea050643e8b385c2dac08641f02/ViT-L-14-336px.pt",
}


class Downsample(eqx.Module):
    avg: eqx.nn.AvgPool2d
    conv: eqx.nn.Conv2d
    bn: BatchNorm

    inference: bool

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        inference: bool,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        self.inference = inference
        _, subkey = jax.random.split(key)
        self.avg = eqx.nn.AvgPool2d(kernel_size=stride, stride=stride)
        self.conv = eqx.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=1,
            stride=1,
            use_bias=False,
            key=subkey,
            dtype=dtype,
        )

        self.bn = BatchNorm(out_channels, axis_name="batch", dtype=dtype)

    def __call__(
        self,
        x: Float[Array, "c_in h w"],
        state: eqx.nn.State,
    ) -> tuple[Float[Array, "c_out*e h/s w/s"], eqx.nn.State]:
        x = self.avg(x)
        x = self.conv(x)
        x, state = self.bn(x, state)

        return x, state


class Bottleneck(eqx.Module):
    conv1: eqx.nn.Conv2d
    bn1: BatchNorm

    conv2: eqx.nn.Conv2d
    bn2: BatchNorm

    conv3: eqx.nn.Conv2d
    bn3: BatchNorm

    downsample: Downsample | None

    avgpool: eqx.nn.AvgPool2d | None

    inference: bool
    expansion: int = eqx.field(static=True, default=4)

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int,
        inference: bool,
        axis_name: str,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        self.inference = inference
        _, *subkeys = jax.random.split(key, 4)

        self.conv1 = eqx.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=1,
            use_bias=False,
            key=subkeys[0],
            dtype=dtype,
        )
        self.bn1 = BatchNorm(
            out_channels, inference=self.inference, axis_name=axis_name, dtype=dtype
        )

        self.conv2 = eqx.nn.Conv2d(
            out_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            use_bias=False,
            key=subkeys[1],
            dtype=dtype,
        )

        self.bn2 = BatchNorm(
            out_channels, inference=inference, axis_name=axis_name, dtype=dtype
        )

        self.avgpool = (
            eqx.nn.AvgPool2d(kernel_size=stride, stride=stride) if stride > 1 else None
        )

        self.conv3 = eqx.nn.Conv2d(
            out_channels,
            out_channels * self.expansion,
            kernel_size=1,
            key=subkeys[2],
            use_bias=False,
            dtype=dtype,
        )

        self.bn3 = BatchNorm(
            out_channels * self.expansion,
            inference=self.inference,
            axis_name=axis_name,
            dtype=dtype,
        )

        self.downsample = None

        if stride > 1 or in_channels != out_channels * Bottleneck.expansion:
            key, subkey = jax.random.split(key)
            self.downsample = Downsample(
                in_channels,
                out_channels * self.expansion,
                inference=self.inference,
                stride=stride,
                key=subkey,
                dtype=dtype,
            )

    def __call__(self, x: Array, state: eqx.nn.State):
        identity = x

        out, state = self.bn1(self.conv1(x), state)
        out = jax.nn.relu(out)

        out, state = self.bn2(self.conv2(out), state)
        out = jax.nn.relu(out)

        if self.avgpool:
            out = self.avgpool(out)
        out, state = self.bn3(self.conv3(out), state)

        if self.downsample is not None:
            identity, state = self.downsample(x, state)

        out += identity
        out = jax.nn.relu(out)
        return out, state


class AttentionPool2d(eqx.Module):
    positional_embedding: Array

    k_proj: eqx.nn.Linear
    q_proj: eqx.nn.Linear
    v_proj: eqx.nn.Linear
    c_proj: eqx.nn.Linear

    inference: bool

    num_heads: int = eqx.field(static=True)

    def __init__(
        self,
        spacial_dim: int,
        embed_dim: int,
        num_heads: int,
        output_dim: int | None,
        inference: bool,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        self.inference = inference
        key, *subkeys = jax.random.split(key, 5)
        self.positional_embedding = (
            jax.random.normal(key, (spacial_dim**2 + 1, embed_dim), dtype=dtype)
            / embed_dim**0.5
        )
        self.k_proj = eqx.nn.Linear(embed_dim, embed_dim, key=subkeys[0], dtype=dtype)
        self.q_proj = eqx.nn.Linear(embed_dim, embed_dim, key=subkeys[1], dtype=dtype)
        self.v_proj = eqx.nn.Linear(embed_dim, embed_dim, key=subkeys[2], dtype=dtype)
        self.c_proj = eqx.nn.Linear(
            embed_dim, output_dim or embed_dim, key=subkeys[3], dtype=dtype
        )
        self.num_heads = num_heads

    def __call__(self, x: Float[Array, "c h w"]):
        c, h, w = x.shape
        x = jnp.einsum("chw->hwc", x).reshape(h * w, c)
        x = jnp.concatenate([jnp.mean(x, axis=0, keepdims=True), x], axis=0)

        x = x + self.positional_embedding

        assert self.q_proj.bias is not None
        assert self.k_proj.bias is not None
        assert self.v_proj.bias is not None
        x, _ = F.multi_head_attention_forward(
            query=x[:1],
            key=x,
            value=x,
            embed_dim_to_check=x.shape[-1],
            num_heads=self.num_heads,
            q_proj_weight=self.q_proj.weight,
            k_proj_weight=self.k_proj.weight,
            v_proj_weight=self.v_proj.weight,
            in_proj_weight=None,
            in_proj_bias=jnp.concatenate(
                [self.q_proj.bias, self.k_proj.bias, self.v_proj.bias]
            ),
            bias_k=None,
            bias_v=None,
            add_zero_attn=False,
            dropout_p=0,
            out_proj_weight=self.c_proj.weight,
            out_proj_bias=self.c_proj.bias,
            use_separate_proj_weight=True,
            inference=self.inference,
            need_weights=False,
        )
        return x.squeeze(0)


class ResNetBlock(eqx.Module):
    blocks: list[Bottleneck]
    inference: bool

    def __call__(self, x, state):
        for block in self.blocks:
            x, state = block(x, state)
        return x, state


class ModifiedResNet(eqx.Module):
    conv1: eqx.nn.Conv2d
    bn1: BatchNorm

    conv2: eqx.nn.Conv2d
    bn2: BatchNorm

    conv3: eqx.nn.Conv2d
    bn3: BatchNorm

    avgpool: eqx.nn.AvgPool2d

    layer1: ResNetBlock
    layer2: ResNetBlock
    layer3: ResNetBlock
    layer4: ResNetBlock

    attnpool: AttentionPool2d

    inference: bool

    _inplanes: int = eqx.field(static=True)
    output_dim: int = eqx.field(static=True)
    input_resolution: int = eqx.field(static=True)

    axis_name: str = eqx.field(static=True)

    def __init__(
        self,
        layers,
        output_dim: int,
        heads: int,
        input_resolution: int,
        width: int,
        inference: bool,
        key: PRNGKeyArray,
        axis_name: str,
        dtype: Any,
    ):
        self.inference = inference
        self.output_dim = output_dim
        self.input_resolution = input_resolution
        self.axis_name = axis_name

        key, *subkeys = jax.random.split(key, 7)

        # the 3-layer stem
        self.conv1 = eqx.nn.Conv2d(
            3,
            width // 2,
            kernel_size=3,
            stride=2,
            padding=1,
            use_bias=False,
            key=subkeys[0],
            dtype=dtype,
        )
        self.bn1 = BatchNorm(
            width // 2, inference=self.inference, dtype=dtype, axis_name=axis_name
        )

        self.conv2 = eqx.nn.Conv2d(
            width // 2,
            width // 2,
            kernel_size=3,
            padding=1,
            use_bias=False,
            key=subkeys[1],
            dtype=dtype,
        )
        self.bn2 = BatchNorm(
            width // 2, inference=self.inference, axis_name=axis_name, dtype=dtype
        )

        self.conv3 = eqx.nn.Conv2d(
            width // 2,
            width,
            kernel_size=3,
            padding=1,
            use_bias=False,
            key=subkeys[2],
            dtype=dtype,
        )
        self.bn3 = BatchNorm(width, axis_name=axis_name, dtype=dtype)
        self.avgpool = eqx.nn.AvgPool2d(2, stride=2)

        key, *subkeys = jax.random.split(key, 7)

        self._inplanes = width
        self.layer1 = self._make_layer(width, layers[0], key=subkeys[0], dtype=dtype)
        self.layer2 = self._make_layer(
            width * 2, layers[1], stride=2, key=subkeys[1], dtype=dtype
        )
        self.layer3 = self._make_layer(
            width * 4, layers[2], stride=2, key=subkeys[2], dtype=dtype
        )
        self.layer4 = self._make_layer(
            width * 8, layers[3], stride=2, key=subkeys[3], dtype=dtype
        )

        embed_dim = width * 32  # the ResNet feature dimension
        self.attnpool = AttentionPool2d(
            input_resolution // 32,
            embed_dim,
            heads,
            output_dim,
            self.inference,
            key=subkeys[4],
            dtype=dtype,
        )

    def _make_layer(
        self, planes, blocks, stride=1, *, key: PRNGKeyArray, dtype: Any | None
    ):
        key, *subkeys = jax.random.split(key, blocks + 1)
        layers = [
            Bottleneck(
                self._inplanes,
                planes,
                stride,
                inference=self.inference,
                axis_name=self.axis_name,
                key=subkeys[0],
                dtype=dtype,
            )
        ]

        self._inplanes = planes * Bottleneck.expansion
        for i in range(1, blocks):
            layers.append(
                Bottleneck(
                    self._inplanes,
                    planes,
                    stride=1,
                    inference=self.inference,
                    axis_name=self.axis_name,
                    key=subkeys[i],
                    dtype=dtype,
                )
            )

        return ResNetBlock(layers, self.inference)

    def __call__(
        self,
        x: Float[Array, "c h w"],
        state: eqx.nn.State | None = None,
    ):
        assert state is not None, "Must give state for ResNet"

        def stem(x, state):
            x, state = self.bn1(self.conv1(x), state)
            x = jax.nn.relu(x)
            x, state = self.bn2(self.conv2(x), state)
            x = jax.nn.relu(x)
            x, state = self.bn3(self.conv3(x), state)
            x = jax.nn.relu(x)
            x = self.avgpool(x)
            return x, state

        x, state = stem(x, state)
        x, state = self.layer1(x, state)
        x, state = self.layer2(x, state)
        x, state = self.layer3(x, state)
        x, state = self.layer4(x, state)
        x = self.attnpool(x)

        return x, state


class ResidualAttentionBlock(eqx.Module):
    attn: MultiheadAttention
    ln_1: eqx.nn.LayerNorm
    c_fc: eqx.nn.Linear
    c_proj: eqx.nn.Linear
    ln_2: eqx.nn.LayerNorm

    inference: bool

    def __init__(
        self, d_model: int, n_head: int, inference: bool, key: PRNGKeyArray, dtype: Any
    ):
        self.inference = inference
        key, *subkeys = jax.random.split(key, 5)
        self.attn = MultiheadAttention(
            d_model, n_head, inference=inference, key=subkeys[0], dtype=dtype
        )
        self.ln_1 = eqx.nn.LayerNorm(d_model, dtype=dtype)
        self.c_fc = eqx.nn.Linear(d_model, d_model * 4, key=subkeys[1], dtype=dtype)
        self.c_proj = eqx.nn.Linear(d_model * 4, d_model, key=subkeys[2], dtype=dtype)
        self.ln_2 = eqx.nn.LayerNorm(d_model, dtype=dtype)

    def _attention(self, x: Array, attn_mask: Array | None = None):
        return self.attn(x, x, x, need_weights=False, attn_mask=attn_mask)[0]

    def __call__(self, x: Array, attn_mask: Array | None = None):
        ln1_out = eqx.filter_vmap(self.ln_1)(x)
        attention_output = self._attention(ln1_out, attn_mask)
        x = x + attention_output
        ln_2_output = eqx.filter_vmap(self.ln_2)(x)
        c_fc_out = eqx.filter_vmap(self.c_fc)(ln_2_output)
        x_gelu = jax.nn.gelu(c_fc_out)
        c_proj_out = eqx.filter_vmap(self.c_proj)(x_gelu)
        x = x + c_proj_out  # Add to x, not to ln_2_output
        return x


class Transformer(eqx.Module):
    resblocks: list[ResidualAttentionBlock]

    inference: bool

    def __init__(
        self,
        width: int,
        layers: int,
        heads: int,
        inference: bool,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        self.inference = inference
        key, *subkeys = jax.random.split(key, layers + 1)
        self.resblocks = [
            ResidualAttentionBlock(
                width, heads, self.inference, key=subkeys[i], dtype=dtype
            )
            for i in range(layers)
        ]

    def __call__(
        self, x: Float[Array, "seq_len embed_dim"], attn_mask: Array | None = None
    ):
        for resblock in self.resblocks:
            x = resblock(x, attn_mask=attn_mask)

        return x


class VisionTransformer(eqx.Module):
    class_embedding: Array
    positional_embedding: Array
    proj: Array
    conv1: eqx.nn.Conv2d
    ln_pre: eqx.nn.LayerNorm
    transformer: Transformer
    ln_post: eqx.nn.LayerNorm

    inference: bool

    def __init__(
        self,
        input_resolution: int,
        patch_size: int,
        width: int,
        layers: int,
        heads: int,
        output_dim: int,
        inference: bool,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        self.inference = inference
        key, *subkeys = jax.random.split(key, 6)
        self.conv1 = eqx.nn.Conv2d(
            in_channels=3,
            out_channels=width,
            kernel_size=patch_size,
            stride=patch_size,
            use_bias=False,
            key=subkeys[0],
            dtype=dtype,
        )

        scale = width**-0.5
        self.class_embedding = (
            jax.random.normal(subkeys[1], (width,), dtype=dtype) * scale
        )
        self.positional_embedding = scale * jax.random.normal(
            subkeys[2], ((input_resolution // patch_size) ** 2 + 1, width), dtype=dtype
        )

        self.ln_pre = eqx.nn.LayerNorm(width, dtype=dtype)

        self.transformer = Transformer(
            width, layers, heads, self.inference, key=subkeys[3], dtype=dtype
        )

        self.ln_post = eqx.nn.LayerNorm(width, dtype=dtype)
        self.proj = scale * jax.random.normal(
            subkeys[4], (width, output_dim), dtype=dtype
        )

    def __call__(
        self,
        x: Float[Array, "c h w"],
        state: eqx.nn.State | None = None,
    ):
        x = self.conv1(x)
        x = x.reshape(x.shape[0], -1)
        x = jnp.transpose(x)

        x = jnp.concatenate(
            [
                self.class_embedding.reshape(1, -1),
                x,
            ],
        )
        x = x + self.positional_embedding
        x = eqx.filter_vmap(self.ln_pre)(x)
        x = self.transformer(x)
        x = self.ln_post(x[0, :])

        if self.proj is not None:
            x = x @ self.proj

        return x, None


class CLIP(eqx.Module):
    positional_embedding: Array
    text_projection: Array
    logit_scale: Array
    visual: ModifiedResNet | VisionTransformer
    transformer: Transformer
    token_embedding: eqx.nn.Embedding
    ln_final: eqx.nn.LayerNorm

    inference: bool

    context_length: int = eqx.field(static=True)
    axis_name: str = eqx.field(static=True)

    def __init__(
        self,
        embed_dim: int,
        # vision
        image_resolution: int,
        vision_layers: tuple[int, int, int, int] | int,
        vision_width: int,
        vision_patch_size: int | None,
        # text
        context_length: int,
        vocab_size: int,
        transformer_width: int,
        transformer_heads: int,
        transformer_layers: int,
        inference: bool,
        axis_name: str,
        key: PRNGKeyArray | None = None,
        dtype: Any | None = None,
    ):
        self.inference = inference
        self.axis_name = axis_name
        if dtype is None:
            dtype = F.default_floating_dtype()
        assert dtype is not None
        self.context_length = context_length
        if key is None:
            # use default key
            key = jax.random.key(42)
        key, *subkeys = jax.random.split(key, 10)

        if isinstance(vision_layers, (tuple, list)):
            vision_heads = vision_width * 32 // 64
            self.visual = ModifiedResNet(
                layers=vision_layers,
                output_dim=embed_dim,
                heads=vision_heads,
                input_resolution=image_resolution,
                width=vision_width,
                inference=self.inference,
                axis_name=self.axis_name,
                key=subkeys[0],
                dtype=dtype,
            )
        else:
            vision_heads = vision_width // 64
            assert vision_patch_size is not None
            self.visual = VisionTransformer(
                input_resolution=image_resolution,
                patch_size=vision_patch_size,
                width=vision_width,
                layers=vision_layers,
                heads=vision_heads,
                output_dim=embed_dim,
                inference=self.inference,
                key=subkeys[0],
                dtype=dtype,
            )
        self.transformer = Transformer(
            width=transformer_width,
            layers=transformer_layers,
            heads=transformer_heads,
            inference=self.inference,
            key=subkeys[1],
            dtype=dtype,
        )
        self.token_embedding = eqx.nn.Embedding(
            weight=jax.random.normal(
                key=subkeys[2], shape=(vocab_size, transformer_width)
            )
            * 0.02,
            dtype=dtype,
        )
        self.positional_embedding = (
            jax.random.normal(
                key=subkeys[3], shape=(context_length, transformer_width), dtype=dtype
            )
            * 0.01
        )
        self.ln_final = eqx.nn.LayerNorm(transformer_width, dtype=dtype)

        self.text_projection = jnp.empty((transformer_width, embed_dim), dtype=dtype)
        self.logit_scale = jnp.ones([], dtype=dtype) * jnp.log(1 / 0.07)

    def initialize_parameters(self, key: PRNGKeyArray):
        # todo
        pass

    def encode_image(self, image, state: eqx.nn.State | None):
        return self.visual(image, state=state)

    def encode_text(self, text: Int[Array, "77"], attn_mask: Array):
        x = eqx.filter_vmap(self.token_embedding)(text)  # [n_ctx, d_model]
        x = x + self.positional_embedding
        x = self.transformer(x, attn_mask=attn_mask)
        x = eqx.filter_vmap(self.ln_final)(x)

        eot_indices = jnp.argmax(text)

        x_at_eot = x[eot_indices]
        x = x_at_eot @ self.text_projection

        return x

    def __call__(
        self,
        image: Float[Array, "c h w"],
        text: Int[Array, "1 77"],
        state: eqx.nn.State | None,
        attn_mask: Array | None = None,
    ):
        image_features, state = self.encode_image(image, state)
        text = text.reshape(-1)
        if attn_mask is None:
            attn_mask = F.build_attention_mask(self.context_length)
        text_features = self.encode_text(text, attn_mask)

        image_norm = jnp.linalg.norm(image_features, axis=-1, keepdims=True)
        text_norm = jnp.linalg.norm(text_features, axis=-1, keepdims=True)
        normalized_image = image_features / image_norm
        normalized_text = text_features / text_norm

        # Calculate similarity with normalized features
        logit_scale = jnp.exp(self.logit_scale)
        logits_per_image = logit_scale * (normalized_image @ normalized_text.T)
        logits_per_text = logits_per_image.T

        return logits_per_image, logits_per_text, state


def _with_weights(
    pytree: PyTree,
    model: Literal[
        "RN50",
        "RN101",
        "RN50x4",
        "RN50x16",
        "RN50x64",
        "ViT-B/32",
        "ViT-B/16",
        "ViT-L/14",
        "ViT-L/14@336px",
    ],
    cache: bool,
    dtype: Any,
):
    clip, state = pytree
    dtype_str = dtype_to_str(dtype)
    if model is not None:
        weights_url = _MODELS.get(model)
        if weights_url is None:
            raise ValueError(f"No model found for {model}")
        jaxonmodels_dir = os.path.expanduser("~/.jaxonmodels/models")
        os.makedirs(jaxonmodels_dir, exist_ok=True)

        if cache:
            if os.path.exists(
                str(
                    Path(jaxonmodels_dir) / f"{model.replace('/', '_')}-{dtype_str}.eqx"
                )
            ):
                return eqx.tree_deserialise_leaves(
                    str(
                        Path(jaxonmodels_dir)
                        / f"{model.replace('/', '_')}-{dtype_str}.eqx"
                    ),
                    (clip, state),
                )
        weights_dir = os.path.expanduser("~/.jaxonmodels/pytorch_weights")
        os.makedirs(weights_dir, exist_ok=True)
        filename = weights_url.split("/")[-1].replace("/", "_")
        weights_file = os.path.join(weights_dir, filename)
        if not os.path.exists(weights_file):
            urlretrieve(weights_url, weights_file)

        import torch

        torch_model = torch.jit.load(weights_file, map_location=torch.device("cpu"))
        weights_dict = {}
        for name, param in torch_model.named_parameters():
            weights_dict[name] = param.clone().detach()
        for name, buffer in torch_model.named_buffers():
            weights_dict[name] = buffer.clone().detach()
        weights_dict["logit_scale"] = torch.Tensor(
            [torch_model.logit_scale.clone().detach()]
        )

        torchfields = state_dict_to_fields(weights_dict)
        torchfields = move_running_fields_to_the_end(torchfields)
        jaxfields, state_indices = pytree_to_fields((clip, state))

        clip, state = convert(
            weights_dict, (clip, state), jaxfields, state_indices, torchfields
        )

        if cache:
            eqx.tree_serialise_leaves(
                str(
                    Path(jaxonmodels_dir) / f"{model.replace('/', '_')}-{dtype_str}.eqx"
                ),
                (clip, state),
            )
    return clip, state


def _clip_resnet50(key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str):
    embed_dim = 1024
    image_resolution = 224
    vision_layers = (3, 4, 6, 3)
    vision_width = 64
    vision_patch_size = None
    context_length = 77
    vocab_size = 49408
    transformer_width = 512
    transformer_heads = 8
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def _clip_resnet101(key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str):
    embed_dim = 512
    image_resolution = 224
    vision_layers = (3, 4, 23, 3)
    vision_width = 64
    vision_patch_size = None
    context_length = 77
    vocab_size = 49408
    transformer_width = 512
    transformer_heads = 8
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def _clip_rn50x4(key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str):
    embed_dim = 640
    image_resolution = 288
    vision_layers = (4, 6, 10, 6)
    vision_width = 80
    vision_patch_size = None
    context_length = 77
    vocab_size = 49408
    transformer_width = 640
    transformer_heads = 10
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def _clip_rn50x16(key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str):
    embed_dim = 768
    image_resolution = 384
    vision_layers = (6, 8, 18, 8)
    vision_width = 96
    vision_patch_size = None
    context_length = 77
    vocab_size = 49408
    transformer_width = 768
    transformer_heads = 12
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def _clip_rn50x64(key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str):
    embed_dim = 1024
    image_resolution = 448
    vision_layers = (3, 15, 36, 10)
    vision_width = 128
    vision_patch_size = None
    context_length = 77
    vocab_size = 49408
    transformer_width = 1024
    transformer_heads = 16
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def _clip_vit_b_16(key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str):
    embed_dim = 512
    image_resolution = 224
    vision_layers = 12
    vision_width = 768
    vision_patch_size = 16
    context_length = 77
    vocab_size = 49408
    transformer_width = 512
    transformer_heads = 8
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def _clip_vit_b_32(key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str):
    embed_dim = 512
    image_resolution = 224
    vision_layers = 12
    vision_width = 768
    vision_patch_size = 32
    context_length = 77
    vocab_size = 49408
    transformer_width = 512
    transformer_heads = 8
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def _clip_vit_l_14(key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str):
    embed_dim = 768
    image_resolution = 224
    vision_layers = 24
    vision_width = 1024
    vision_patch_size = 14
    context_length = 77
    vocab_size = 49408
    transformer_width = 768
    transformer_heads = 12
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def _clip_vit_l_14_336px(
    key: PRNGKeyArray, dtype: Any, inference: bool, axis_name: str
):
    embed_dim = 768
    image_resolution = 336
    vision_layers = 24
    vision_width = 1024
    vision_patch_size = 14
    context_length = 77
    vocab_size = 49408
    transformer_width = 768
    transformer_heads = 12
    transformer_layers = 12

    clip, state = eqx.nn.make_with_state(CLIP)(
        embed_dim=embed_dim,
        image_resolution=image_resolution,
        vision_layers=vision_layers,
        vision_width=vision_width,
        vision_patch_size=vision_patch_size,
        context_length=context_length,
        vocab_size=vocab_size,
        transformer_width=transformer_width,
        transformer_heads=transformer_heads,
        transformer_layers=transformer_layers,
        inference=inference,
        axis_name=axis_name,
        key=key,
        dtype=dtype,
    )

    return clip, state


def load_clip(
    model: Literal[
        "RN50",
        "RN101",
        "RN50x4",
        "RN50x16",
        "RN50x64",
        "ViT-B/32",
        "ViT-B/16",
        "ViT-L/14",
        "ViT-L/14@336px",
    ],
    with_weights: bool = False,
    cache: bool = True,
    inference: bool = True,
    axis_name: str = "batch",
    *,
    key: PRNGKeyArray | None = None,
    dtype: Any | None = None,
):
    """Loads a CLIP model.

    Args:
        model: The name of the CLIP model variant to load.
        with_weights: Whether to load the pretrained weights from OpenAI.
        cache: Whether to cache the downloaded weights and the converted model.
        inference: Whether to initialize the model in inference mode.
                   (e.g., for BatchNorm)
                   Defaults to True.
        axis_name: The name of the batch axis used for BatchNorm synchronization.
                   Defaults to "batch".
        key: A JAX PRNG key for initializing parameters if `with_weights=False`.
        dtype: The data type to use for the model parameters. Defaults to the
               `jaxonmodels` default floating-point type.

    Returns:
        A tuple `(clip_model, state)` where `clip_model` is the CLIP model
        and `state` is the initial state (e.g., for BatchNorm statistics).
    """
    if dtype is None:
        dtype = F.default_floating_dtype()
    assert dtype is not None
    if key is None:
        key = jax.random.key(42)
    clip, state = None, None

    match model:
        case "RN50":
            clip, state = _clip_resnet50(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )
        case "RN101":
            clip, state = _clip_resnet101(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )
        case "RN50x4":
            clip, state = _clip_rn50x4(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )
        case "RN50x16":
            clip, state = _clip_rn50x16(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )
        case "RN50x64":
            clip, state = _clip_rn50x64(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )
        case "ViT-B/32":
            clip, state = _clip_vit_b_32(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )
        case "ViT-B/16":
            clip, state = _clip_vit_b_16(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )
        case "ViT-L/14":
            clip, state = _clip_vit_l_14(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )
        case "ViT-L/14@336px":
            clip, state = _clip_vit_l_14_336px(
                key=key, dtype=dtype, inference=inference, axis_name=axis_name
            )

    if clip is None or state is None:
        raise ValueError(f"Unrecognised model passed: {model}")

    if with_weights:
        clip, state = _with_weights((clip, state), model, cache, dtype=dtype)

    assert clip is not None
    assert state is not None
    return clip, state



import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any
from jaxonlayers.layers import MultiheadAttention
from jaxtyping import Array, Float, PRNGKeyArray

from jaxonmodels.functions.utils import default_floating_dtype


class ResidualAttentionBlock(eqx.Module):
    attn: MultiheadAttention
    ln_1: eqx.nn.LayerNorm
    c_fc: eqx.nn.Linear
    c_proj: eqx.nn.Linear
    ln_2: eqx.nn.LayerNorm

    def __init__(self, d_model: int, n_head: int, *, key: PRNGKeyArray, dtype: Any):
        key, *subkeys = jax.random.split(key, 5)
        self.attn = MultiheadAttention(d_model, n_head, key=subkeys[0], dtype=dtype)
        self.ln_1 = eqx.nn.LayerNorm(d_model, dtype=dtype)
        self.c_fc = eqx.nn.Linear(d_model, d_model * 4, key=subkeys[1], dtype=dtype)
        self.c_proj = eqx.nn.Linear(d_model * 4, d_model, key=subkeys[2], dtype=dtype)
        self.ln_2 = eqx.nn.LayerNorm(d_model, dtype=dtype)

    def _attention(self, x: Array, attn_mask: Array | None = None):
        return self.attn(x, x, x, need_weights=False, attn_mask=attn_mask)[0]

    def __call__(self, x: Array, attn_mask: Array | None = None):
        ln1_out = eqx.filter_vmap(self.ln_1)(x)
        attention_output = self._attention(ln1_out, attn_mask)
        x = x + attention_output
        ln_2_output = eqx.filter_vmap(self.ln_2)(x)
        c_fc_out = eqx.filter_vmap(self.c_fc)(ln_2_output)
        x_gelu = jax.nn.gelu(c_fc_out)
        c_proj_out = eqx.filter_vmap(self.c_proj)(x_gelu)
        x = x + c_proj_out  # Add to x, not to ln_2_output
        return x


class Transformer(eqx.Module):
    resblocks: list[ResidualAttentionBlock]

    def __init__(
        self, width: int, layers: int, heads: int, *, key: PRNGKeyArray, dtype: Any
    ):
        key, *subkeys = jax.random.split(key, layers + 1)
        self.resblocks = [
            ResidualAttentionBlock(width, heads, key=subkeys[i], dtype=dtype)
            for i in range(layers)
        ]

    def __call__(
        self, x: Float[Array, "seq_len embed_dim"], attn_mask: Array | None = None
    ):
        for resblock in self.resblocks:
            x = resblock(x, attn_mask=attn_mask)

        return x


class VisionTransformer(eqx.Module):
    class_embedding: Array
    positional_embedding: Array
    proj: Array
    conv1: eqx.nn.Conv2d
    ln_pre: eqx.nn.LayerNorm
    transformer: Transformer
    ln_post: eqx.nn.LayerNorm

    def __init__(
        self,
        input_resolution: int,
        patch_size: int,
        width: int,
        layers: int,
        heads: int,
        output_dim: int,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        key, *subkeys = jax.random.split(key, 6)
        self.conv1 = eqx.nn.Conv2d(
            in_channels=3,
            out_channels=width,
            kernel_size=patch_size,
            stride=patch_size,
            use_bias=False,
            key=subkeys[0],
        )

        scale = width**-0.5
        self.class_embedding = (
            jax.random.normal(subkeys[1], (width,), dtype=dtype) * scale
        )
        self.positional_embedding = scale * jax.random.normal(
            subkeys[2], ((input_resolution // patch_size) ** 2 + 1, width), dtype=dtype
        )

        self.ln_pre = eqx.nn.LayerNorm(width)

        self.transformer = Transformer(
            width, layers, heads, key=subkeys[3], dtype=dtype
        )

        self.ln_post = eqx.nn.LayerNorm(width, dtype=dtype)
        self.proj = scale * jax.random.normal(
            subkeys[4], (width, output_dim), dtype=dtype
        )

    def __call__(
        self,
        x: Float[Array, "c h w"],
        state: eqx.nn.State | None = None,
        *,
        inference: bool = False,
        key: PRNGKeyArray | None = None,
    ):
        x = self.conv1(x)
        x = x.reshape(x.shape[0], -1)
        x = jnp.transpose(x)

        x = jnp.concatenate(
            [
                self.class_embedding.reshape(1, -1),
                x,
            ],
        )
        x = x + self.positional_embedding
        x = eqx.filter_vmap(self.ln_pre)(x)
        x = self.transformer(x)
        x = self.ln_post(x[0, :])

        if self.proj is not None:
            x = x @ self.proj

        return x, state



import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from esm.layers.attention import MultiHeadAttention
from esm.layers.blocks import UnifiedTransformerBlock as TorchBlock
from esm.layers.geom_attention import (
    GeometricReasoningOriginalImpl as TorchGeomAttn,
)
from esm.layers.transformer_stack import TransformerStack as TorchTransformerStack
from esm.models.esm3 import ESM3 as TorchESM3
from esm.models.esm3 import EncodeInputs as TorchEncodeInputs
from esm.models.esm3 import OutputHeads as TorchOutputHeads
from esm.models.esmc import ESMC as TorchESMC
from esm.tokenization import EsmSequenceTokenizer
from esm.utils.structure.affine3d import (
    Affine3D as TorchAffine3D,
)
from esm.utils.structure.affine3d import (
    RotationMatrix as TorchRotationMatrix,
)
from statedict2pytree.converter import autoconvert

from jaxonmodels.models.esm import ESM3 as JaxESM3
from jaxonmodels.models.esm import ESMC as JaxESMC
from jaxonmodels.models.esm import VQVAE_SPECIAL_TOKENS, ESMMultiHeadAttention
from jaxonmodels.models.esm import (
    Affine3D as JaxAffine3D,
)
from jaxonmodels.models.esm import Dim6RotStructureHead as JaxDim6Rot
from jaxonmodels.models.esm import EncodeInputs as JaxEncodeInputs
from jaxonmodels.models.esm import (
    GeometricReasoningOriginalImpl as JaxGeomAttn,
)
from jaxonmodels.models.esm import OutputHeads as JaxOutputHeads
from jaxonmodels.models.esm import PairwisePredictionHead as JaxPairwisePredictionHead
from jaxonmodels.models.esm import (
    RotationMatrix as JaxRotationMatrix,
)
from jaxonmodels.models.esm import StructureTokenDecoder as JaxStructureTokenDecoder
from jaxonmodels.models.esm import UnifiedTransformerBlock as JaxBlock


@pytest.mark.parametrize(
    "d_model, n_heads, seq_len, bias, qk_layernorm",
    [(96, 4, 7, True, True), (960, 15, 7, False, True), (1152, 18, 7, False, True)],
)
def test_mha_implementations(
    d_model: int, n_heads: int, seq_len, bias: bool, qk_layernorm: bool
):
    np.random.seed(42)
    x = np.random.uniform(size=(1, seq_len, d_model))
    x = np.array(x, dtype=np.float32)
    seq_id = np.array([[True for _ in range(seq_len)]])

    torch_mha = MultiHeadAttention(d_model, n_heads, bias, qk_layernorm=qk_layernorm)
    state_dict = torch_mha.state_dict()
    torch_output = torch_mha(torch.from_numpy(x), torch.from_numpy(seq_id))

    key = jax.random.key(42)
    jax_mha = ESMMultiHeadAttention(d_model, n_heads, bias, key=key)
    jax_mha = autoconvert(jax_mha, state_dict)
    x = jnp.array(x, dtype=jnp.float32)
    seq_id = jnp.array([[True for _ in range(seq_len)]])

    jax_output = eqx.filter_vmap(jax_mha)(x, seq_id)

    torch_output = torch_output.detach().numpy()
    jax_output = np.array(jax_output)

    assert np.allclose(torch_output, jax_output, atol=1e-5)


def _make_affines(batch, seq_len, seed=42):
    np.random.seed(seed)
    trans = np.random.randn(batch, seq_len, 3).astype(np.float32)
    rots = np.random.randn(batch, seq_len, 3, 3).astype(np.float32)

    torch_affine = TorchAffine3D(
        torch.from_numpy(trans),
        TorchRotationMatrix(torch.from_numpy(rots)),
    )
    jax_affine = JaxAffine3D(
        jnp.array(trans),
        JaxRotationMatrix(jnp.array(rots)),
    )
    return torch_affine, jax_affine


@pytest.mark.parametrize(
    "c_s, v_heads, seq_len, mask_and_zero, use_seq_id",
    [
        (64, 4, 7, True, False),
        (64, 4, 7, False, True),
        (128, 8, 11, True, True),
        (128, 8, 11, False, False),
    ],
)
def test_geom_attention(c_s, v_heads, seq_len, mask_and_zero, use_seq_id):
    batch = 2
    np.random.seed(42)

    s = np.random.randn(batch, seq_len, c_s).astype(np.float32)
    affine_mask = np.ones((batch, seq_len), dtype=bool)
    affine_mask[:, -1] = False
    chain_id = np.zeros((batch, seq_len), dtype=np.int64)

    if use_seq_id:
        sequence_id = np.zeros((batch, seq_len), dtype=np.int64)
        torch_seq_id = torch.from_numpy(sequence_id)
        jax_seq_id = jnp.array(sequence_id)
    else:
        torch_seq_id = None
        jax_seq_id = None

    torch_affine, jax_affine = _make_affines(batch, seq_len)

    torch_geom = TorchGeomAttn(
        c_s=c_s,
        v_heads=v_heads,
        mask_and_zero_frameless=mask_and_zero,
        bias=False,
    )
    state_dict = torch_geom.state_dict()

    torch_output = (
        torch_geom(
            torch.from_numpy(s),
            torch_affine,
            torch.from_numpy(affine_mask),
            torch_seq_id,
            torch.from_numpy(chain_id),
        )
        .detach()
        .numpy()
    )

    key = jax.random.key(42)
    jax_geom = JaxGeomAttn(
        c_s=c_s,
        v_heads=v_heads,
        mask_and_zero_frameless=mask_and_zero,
        bias=False,
        key=key,
    )
    jax_geom = autoconvert(jax_geom, state_dict)

    def call_single(s_i, affine_i, affine_mask_i, seq_id_i, chain_id_i):
        return jax_geom(s_i, affine_i, affine_mask_i, seq_id_i, chain_id_i)

    jax_s = jnp.array(s)
    jax_affine_mask = jnp.array(affine_mask)
    jax_chain_id = jnp.array(chain_id)

    if jax_seq_id is not None:
        jax_output = np.array(
            eqx.filter_vmap(call_single)(
                jax_s,
                jax_affine,
                jax_affine_mask,
                jax_seq_id,
                jax_chain_id,
            )
        )
    else:
        jax_output = np.array(
            eqx.filter_vmap(call_single, in_axes=(0, 0, 0, None, 0))(
                jax_s,
                jax_affine,
                jax_affine_mask,
                None,
                jax_chain_id,
            )
        )

    assert np.allclose(torch_output, jax_output, atol=1e-4)


def _make_block_inputs(batch, seq_len, d_model, use_frames, use_seq_id, seed=42):
    np.random.seed(seed)
    x = np.random.randn(batch, seq_len, d_model).astype(np.float32)
    chain_id = np.zeros((batch, seq_len), dtype=np.int64)

    if use_seq_id:
        sequence_id = np.zeros((batch, seq_len), dtype=np.int64)
    else:
        sequence_id = None

    if use_frames:
        frames_mask = np.ones((batch, seq_len), dtype=bool)
        frames_mask[:, -1] = False
        trans = np.random.randn(batch, seq_len, 3).astype(np.float32)
        rots = np.random.randn(batch, seq_len, 3, 3).astype(np.float32)
        torch_frames = TorchAffine3D(
            torch.from_numpy(trans), TorchRotationMatrix(torch.from_numpy(rots))
        )
        jax_trans = jnp.array(trans)
        jax_rots = jnp.array(rots)
    else:
        frames_mask = np.ones((batch, seq_len), dtype=bool)
        torch_frames = None
        jax_trans = None
        jax_rots = None

    return x, sequence_id, frames_mask, chain_id, torch_frames, jax_trans, jax_rots


@pytest.mark.parametrize(
    "d_model, n_heads, use_geom, v_heads, bias, expansion_ratio, scaling, qk_ln, ffn_type, use_frames, use_seq_id",  # noqa
    [
        (64, 4, False, None, False, 8 / 3, 1.0, True, "swiglu", False, True),
        (64, 4, True, 8, False, 8 / 3, 1.1547, True, "swiglu", True, False),
        (64, 4, False, None, True, 4.0, 1.0, False, "gelu", False, True),
        (96, 6, True, 4, False, 8 / 3, 1.1547, True, "swiglu", True, True),
    ],
)
def test_unified_transformer_block(
    d_model,
    n_heads,
    use_geom,
    v_heads,
    bias,
    expansion_ratio,
    scaling,
    qk_ln,
    ffn_type,
    use_frames,
    use_seq_id,
):
    batch = 2
    seq_len = 7

    x, sequence_id, frames_mask, chain_id, torch_frames, jax_trans, jax_rots = (
        _make_block_inputs(batch, seq_len, d_model, use_frames, use_seq_id)
    )

    torch_block = TorchBlock(
        d_model=d_model,
        n_heads=n_heads,
        use_geom_attn=use_geom,
        v_heads=v_heads,
        bias=bias,
        expansion_ratio=expansion_ratio,
        residue_scaling_factor=scaling,
        qk_layernorm=qk_ln,
        ffn_type=ffn_type,
    )
    state_dict = torch_block.state_dict()

    torch_output = (
        torch_block(
            torch.from_numpy(x),
            torch.from_numpy(sequence_id) if sequence_id is not None else None,
            torch_frames,
            torch.from_numpy(frames_mask),
            torch.from_numpy(chain_id),
        )
        .detach()
        .numpy()
    )

    key = jax.random.key(42)
    jax_block = JaxBlock(
        d_model=d_model,
        n_heads=n_heads,
        use_geom_attn=use_geom,
        v_heads=v_heads,
        bias=bias,
        expansion_ratio=expansion_ratio,
        residue_scaling_factor=scaling,
        qk_layernorm=qk_ln,
        ffn_type=ffn_type,
        key=key,
    )
    jax_block = autoconvert(jax_block, state_dict)

    jax_x = jnp.array(x)
    jax_frames_mask = jnp.array(frames_mask)
    jax_chain_id = jnp.array(chain_id)
    jax_seq_id = jnp.array(sequence_id) if sequence_id is not None else None

    def call_single(x_i, trans_i, rots_i, frames_mask_i, seq_id_i, chain_id_i):
        if trans_i is not None:
            frames_i = JaxAffine3D(trans_i, JaxRotationMatrix(rots_i))
        else:
            frames_i = None
        return jax_block(x_i, seq_id_i, frames_i, frames_mask_i, chain_id_i)

    if use_frames:
        if jax_seq_id is not None:
            jax_output = np.array(
                eqx.filter_vmap(call_single)(
                    jax_x,
                    jax_trans,
                    jax_rots,
                    jax_frames_mask,
                    jax_seq_id,
                    jax_chain_id,
                )
            )
        else:
            jax_output = np.array(
                eqx.filter_vmap(call_single, in_axes=(0, 0, 0, 0, None, 0))(
                    jax_x,
                    jax_trans,
                    jax_rots,
                    jax_frames_mask,
                    None,
                    jax_chain_id,
                )
            )
    else:
        if jax_seq_id is not None:
            jax_output = np.array(
                eqx.filter_vmap(call_single, in_axes=(0, None, None, 0, 0, 0))(
                    jax_x,
                    None,
                    None,
                    jax_frames_mask,
                    jax_seq_id,
                    jax_chain_id,
                )
            )
        else:
            jax_output = np.array(
                eqx.filter_vmap(call_single, in_axes=(0, None, None, 0, None, 0))(
                    jax_x,
                    None,
                    None,
                    jax_frames_mask,
                    None,
                    jax_chain_id,
                )
            )

    assert np.allclose(torch_output, jax_output, atol=1e-3)


def _make_stack_inputs(batch, seq_len, d_model, use_frames, use_seq_id, seed=42):
    np.random.seed(seed)
    x = np.random.randn(batch, seq_len, d_model).astype(np.float32)
    chain_id = np.zeros((batch, seq_len), dtype=np.int64)

    if use_seq_id:
        sequence_id = np.zeros((batch, seq_len), dtype=np.int64)
    else:
        sequence_id = None

    if use_frames:
        frames_mask = np.ones((batch, seq_len), dtype=bool)
        frames_mask[:, -1] = False
        trans = np.random.randn(batch, seq_len, 3).astype(np.float32)
        rots = np.random.randn(batch, seq_len, 3, 3).astype(np.float32)
        torch_frames = TorchAffine3D(
            torch.from_numpy(trans), TorchRotationMatrix(torch.from_numpy(rots))
        )
        jax_trans = jnp.array(trans)
        jax_rots = jnp.array(rots)
    else:
        frames_mask = np.ones((batch, seq_len), dtype=bool)
        torch_frames = None
        jax_trans = None
        jax_rots = None

    return x, sequence_id, frames_mask, chain_id, torch_frames, jax_trans, jax_rots


from jaxonmodels.models.esm import TransformerStack as JaxTransformerStack


@pytest.mark.parametrize(
    "d_model, n_heads, v_heads, n_layers, n_layers_geom, use_frames, use_seq_id",
    [
        (64, 4, None, 3, 0, False, True),
        (64, 4, 4, 3, 1, True, False),
    ],
)
def test_transformer_stack(
    d_model,
    n_heads,
    v_heads,
    n_layers,
    n_layers_geom,
    use_frames,
    use_seq_id,
):
    batch = 2
    seq_len = 7

    x, sequence_id, frames_mask, chain_id, torch_frames, jax_trans, jax_rots = (
        _make_stack_inputs(batch, seq_len, d_model, use_frames, use_seq_id)
    )

    torch_stack = TorchTransformerStack(
        d_model=d_model,
        n_heads=n_heads,
        v_heads=v_heads,
        n_layers=n_layers,
        n_layers_geom=n_layers_geom,
    )
    state_dict = torch_stack.state_dict()

    torch_x, torch_embed, torch_hiddens = torch_stack(
        torch.from_numpy(x),
        torch.from_numpy(sequence_id) if sequence_id is not None else None,
        torch_frames,
        torch.from_numpy(frames_mask),
        torch.from_numpy(chain_id),
    )
    torch_x = torch_x.detach().numpy()
    torch_embed = torch_embed.detach().numpy()

    key = jax.random.key(42)
    jax_stack = JaxTransformerStack(
        d_model=d_model,
        n_heads=n_heads,
        v_heads=v_heads,
        n_layers=n_layers,
        n_layers_geom=n_layers_geom,
        key=key,
    )
    jax_stack = autoconvert(jax_stack, state_dict)

    jax_x_in = jnp.array(x)
    jax_frames_mask = jnp.array(frames_mask)
    jax_chain_id = jnp.array(chain_id)
    jax_seq_id = jnp.array(sequence_id) if sequence_id is not None else None

    def call_single(x_i, trans_i, rots_i, frames_mask_i, seq_id_i, chain_id_i):
        if trans_i is not None:
            frames_i = JaxAffine3D(trans_i, JaxRotationMatrix(rots_i))
        else:
            frames_i = None
        return jax_stack(x_i, seq_id_i, frames_i, frames_mask_i, chain_id_i)

    if use_frames:
        if jax_seq_id is not None:
            jax_out = eqx.filter_vmap(call_single)(
                jax_x_in,
                jax_trans,
                jax_rots,
                jax_frames_mask,
                jax_seq_id,
                jax_chain_id,
            )
        else:
            jax_out = eqx.filter_vmap(call_single, in_axes=(0, 0, 0, 0, None, 0))(
                jax_x_in,
                jax_trans,
                jax_rots,
                jax_frames_mask,
                None,
                jax_chain_id,
            )
    else:
        if jax_seq_id is not None:
            jax_out = eqx.filter_vmap(call_single, in_axes=(0, None, None, 0, 0, 0))(
                jax_x_in,
                None,
                None,
                jax_frames_mask,
                jax_seq_id,
                jax_chain_id,
            )
        else:
            jax_out = eqx.filter_vmap(call_single, in_axes=(0, None, None, 0, None, 0))(
                jax_x_in,
                None,
                None,
                jax_frames_mask,
                None,
                jax_chain_id,
            )

    jax_x_out, jax_embed, jax_hiddens = jax_out

    assert np.allclose(torch_x, np.array(jax_x_out), atol=1e-3)
    assert np.allclose(torch_embed, np.array(jax_embed), atol=1e-3)
    assert len(torch_hiddens) == len(jax_hiddens)
    for i in range(len(torch_hiddens)):
        assert np.allclose(
            torch_hiddens[i].detach().numpy(), np.array(jax_hiddens[i]), atol=1e-3
        )


@pytest.mark.parametrize(
    "d_model, n_heads, n_layers, seq_len",
    [
        (64, 4, 2, 7),
        (96, 6, 3, 11),
    ],
)
def test_esmc(d_model, n_heads, n_layers, seq_len):
    batch = 2
    np.random.seed(42)

    tokenizer = EsmSequenceTokenizer()
    tokens = np.random.randint(0, 33, size=(batch, seq_len)).astype(np.int64)
    tokens[:, 0] = 0
    tokens[:, -1] = 2

    torch_model = TorchESMC(d_model, n_heads, n_layers, tokenizer)
    state_dict = torch_model.state_dict()

    with torch.no_grad():
        output = torch_model(torch.from_numpy(tokens))
        torch_logits, torch_embed, torch_hiddens = (
            output.sequence_logits,
            output.embeddings,
            output.hidden_states,
        )
    torch_logits = torch_logits.numpy()
    torch_embed = torch_embed.numpy()
    torch_hiddens = torch_hiddens.numpy()

    key = jax.random.key(42)
    jax_model = JaxESMC(d_model, n_heads, n_layers, key=key)
    jax_model = autoconvert(jax_model, state_dict)

    jax_tokens = jnp.array(tokens)

    def call_single(tokens_i):
        return jax_model(tokens_i)

    jax_logits, jax_embed, jax_hiddens = eqx.filter_vmap(call_single)(jax_tokens)
    jax_hiddens = np.array(jax_hiddens).transpose(1, 0, 2, 3)
    assert np.allclose(torch_hiddens, jax_hiddens, atol=1e-3)

    assert np.allclose(torch_logits, np.array(jax_logits), atol=1e-3)
    assert np.allclose(torch_embed, np.array(jax_embed), atol=1e-3)


@pytest.mark.parametrize(
    "d_model, seq_len",
    [
        (64, 7),
        (128, 11),
    ],
)
def test_encode_inputs(d_model, seq_len):
    batch = 2
    np.random.seed(42)

    sequence_tokens = np.random.randint(0, 64, size=(batch, seq_len)).astype(np.int64)
    structure_tokens = np.random.randint(0, 4096, size=(batch, seq_len)).astype(
        np.int64
    )
    ss8_tokens = np.random.randint(0, 11, size=(batch, seq_len)).astype(np.int64)
    sasa_tokens = np.random.randint(0, 19, size=(batch, seq_len)).astype(np.int64)
    function_tokens = np.random.randint(0, 260, size=(batch, seq_len, 8)).astype(
        np.int64
    )
    residue_annotation_tokens = np.random.randint(
        0, 1478, size=(batch, seq_len, 16)
    ).astype(np.int64)
    average_plddt = np.float32(1.0)
    per_res_plddt = np.random.rand(batch, seq_len).astype(np.float32)

    torch_encoder = TorchEncodeInputs(d_model)
    state_dict = torch_encoder.state_dict()

    torch_output = (
        torch_encoder(
            torch.from_numpy(sequence_tokens),
            torch.from_numpy(structure_tokens),
            torch.tensor(average_plddt),
            torch.from_numpy(per_res_plddt),
            torch.from_numpy(ss8_tokens),
            torch.from_numpy(sasa_tokens),
            torch.from_numpy(function_tokens),
            torch.from_numpy(residue_annotation_tokens),
        )
        .detach()
        .numpy()
    )

    key = jax.random.key(42)
    jax_encoder = JaxEncodeInputs(d_model, key=key)
    jax_encoder = autoconvert(jax_encoder, state_dict)

    def call_single(seq, struct, res_plddt, ss8, sasa, func, residue):
        return jax_encoder(
            seq, struct, average_plddt, res_plddt, ss8, sasa, func, residue
        )

    jax_output = np.array(
        eqx.filter_vmap(call_single)(
            jnp.array(sequence_tokens),
            jnp.array(structure_tokens),
            jnp.array(per_res_plddt),
            jnp.array(ss8_tokens),
            jnp.array(sasa_tokens),
            jnp.array(function_tokens),
            jnp.array(residue_annotation_tokens),
        )
    )

    assert np.allclose(torch_output, jax_output, atol=1e-4)


@pytest.mark.parametrize(
    "d_model, seq_len",
    [
        (64, 7),
        (128, 11),
    ],
)
def test_output_heads(d_model, seq_len):
    batch = 2
    np.random.seed(42)

    x = np.random.randn(batch, seq_len, d_model).astype(np.float32)
    embed = np.random.randn(batch, seq_len, d_model).astype(np.float32)

    torch_heads = TorchOutputHeads(d_model)
    state_dict = torch_heads.state_dict()

    torch_out = torch_heads(torch.from_numpy(x), torch.from_numpy(embed))
    torch_results = [
        torch_out.sequence_logits.detach().numpy(),
        torch_out.structure_logits.detach().numpy(),
        torch_out.secondary_structure_logits.detach().numpy(),
        torch_out.sasa_logits.detach().numpy(),
        torch_out.function_logits.detach().numpy(),
        torch_out.residue_logits.detach().numpy(),
        torch_out.embeddings.detach().numpy(),
    ]

    key = jax.random.key(42)
    jax_heads = JaxOutputHeads(d_model, key=key)
    jax_heads = autoconvert(jax_heads, state_dict)

    def call_single(x_i, embed_i):
        return jax_heads(x_i, embed_i)

    jax_out = eqx.filter_vmap(call_single)(jnp.array(x), jnp.array(embed))

    for i in range(len(torch_results)):
        assert np.allclose(torch_results[i], np.array(jax_out[i]), atol=1e-3), (
            f"Mismatch at output index {i}"
        )


class _MockSeqTok:
    mask_token_id = 32


class _MockTokenizers:
    sequence = _MockSeqTok()


@pytest.mark.parametrize(
    "d_model, n_heads, v_heads, n_layers, seq_len",
    [
        (64, 4, 4, 2, 7),
        (96, 6, 4, 3, 11),
    ],
)
def test_esm3_forward(d_model, n_heads, v_heads, n_layers, seq_len):
    batch = 2
    np.random.seed(42)

    sequence_tokens = np.random.randint(3, 31, size=(batch, seq_len)).astype(np.int64)
    sequence_tokens[:, 0] = 0
    sequence_tokens[:, -1] = 2
    structure_tokens = np.random.randint(0, 4096, size=(batch, seq_len)).astype(
        np.int64
    )
    ss8_tokens = np.full((batch, seq_len), 8, dtype=np.int64)
    sasa_tokens = np.full((batch, seq_len), 16, dtype=np.int64)
    function_tokens = np.full((batch, seq_len, 8), 0, dtype=np.int64)
    residue_annotation_tokens = np.full((batch, seq_len, 16), 0, dtype=np.int64)
    average_plddt = np.float32(1.0)
    per_res_plddt = np.zeros((batch, seq_len), dtype=np.float32)
    structure_coords = np.full((batch, seq_len, 3, 3), np.nan, dtype=np.float32)
    chain_id = np.zeros((batch, seq_len), dtype=np.int64)

    torch_model = TorchESM3(
        d_model,
        n_heads,
        v_heads,
        n_layers,
        structure_encoder_fn=None,  # ty: ignore[invalid-argument-type]
        structure_decoder_fn=None,  # ty: ignore[invalid-argument-type]
        function_decoder_fn=None,  # ty: ignore[invalid-argument-type]
        tokenizers=_MockTokenizers(),  # ty: ignore[invalid-argument-type]
    )
    state_dict = torch_model.state_dict()

    with torch.no_grad():
        torch_out = torch_model(
            sequence_tokens=torch.from_numpy(sequence_tokens),
            structure_tokens=torch.from_numpy(structure_tokens),
            ss8_tokens=torch.from_numpy(ss8_tokens),
            sasa_tokens=torch.from_numpy(sasa_tokens),
            function_tokens=torch.from_numpy(function_tokens),
            residue_annotation_tokens=torch.from_numpy(residue_annotation_tokens),
            average_plddt=torch.tensor(average_plddt),
            per_res_plddt=torch.from_numpy(per_res_plddt),
            structure_coords=torch.from_numpy(structure_coords),
            chain_id=torch.from_numpy(chain_id),
        )

    torch_seq_logits = torch_out.sequence_logits.numpy()
    torch_struct_logits = torch_out.structure_logits.numpy()
    torch_embed = torch_out.embeddings.numpy()

    key = jax.random.key(42)
    jax_model = JaxESM3(d_model, n_heads, v_heads, n_layers, key=key)
    jax_model = autoconvert(jax_model, state_dict)

    def call_single(seq, struct, ss8, sasa, func, residue, plddt, coords, cid):
        return jax_model(
            sequence_tokens=seq,
            structure_tokens=struct,
            ss8_tokens=ss8,
            sasa_tokens=sasa,
            function_tokens=func,
            residue_annotation_tokens=residue,
            average_plddt=average_plddt,
            per_res_plddt=plddt,
            structure_coords=coords,
            chain_id=cid,
        )

    jax_out = eqx.filter_vmap(call_single)(
        jnp.array(sequence_tokens),
        jnp.array(structure_tokens),
        jnp.array(ss8_tokens),
        jnp.array(sasa_tokens),
        jnp.array(function_tokens),
        jnp.array(residue_annotation_tokens),
        jnp.array(per_res_plddt),
        jnp.array(structure_coords),
        jnp.array(chain_id),
    )

    jax_seq_logits = np.array(jax_out[0])
    jax_struct_logits = np.array(jax_out[1])
    jax_embed = np.array(jax_out[6])

    assert np.allclose(torch_seq_logits, jax_seq_logits, atol=1e-3)
    assert np.allclose(torch_struct_logits, jax_struct_logits, atol=1e-3)
    assert np.allclose(torch_embed, jax_embed, atol=1e-3)


def test_esm3_forward_debug():
    d_model, n_heads, v_heads, n_layers, seq_len = 64, 4, 4, 2, 7
    batch = 2
    np.random.seed(42)

    sequence_tokens = np.random.randint(3, 31, size=(batch, seq_len)).astype(np.int64)
    sequence_tokens[:, 0] = 0
    sequence_tokens[:, -1] = 2
    structure_tokens = np.random.randint(0, 4096, size=(batch, seq_len)).astype(
        np.int64
    )
    ss8_tokens = np.full((batch, seq_len), 8, dtype=np.int64)
    sasa_tokens = np.full((batch, seq_len), 16, dtype=np.int64)
    function_tokens = np.full((batch, seq_len, 8), 0, dtype=np.int64)
    residue_annotation_tokens = np.full((batch, seq_len, 16), 0, dtype=np.int64)
    average_plddt = np.float32(1.0)
    per_res_plddt = np.zeros((batch, seq_len), dtype=np.float32)
    structure_coords = np.full((batch, seq_len, 3, 3), np.nan, dtype=np.float32)
    chain_id = np.zeros((batch, seq_len), dtype=np.int64)

    torch_model = TorchESM3(
        d_model,
        n_heads,
        v_heads,
        n_layers,
        structure_encoder_fn=None,  # ty: ignore[invalid-argument-type]
        structure_decoder_fn=None,  # ty: ignore[invalid-argument-type]
        function_decoder_fn=None,  # ty: ignore[invalid-argument-type]
        tokenizers=_MockTokenizers(),  # ty: ignore[invalid-argument-type]
    )
    state_dict = torch_model.state_dict()

    key = jax.random.key(42)
    jax_model = JaxESM3(d_model, n_heads, v_heads, n_layers, key=key)
    jax_model = autoconvert(jax_model, state_dict)

    sq_t = torch.from_numpy(sequence_tokens)
    st_t = torch.from_numpy(structure_tokens).clone()
    st_t = st_t.masked_fill(st_t == -1, 4096)
    st_t = st_t.masked_fill(sq_t == 0, 4098)
    st_t = st_t.masked_fill(sq_t == 1, 4097)
    st_t = st_t.masked_fill(sq_t == 2, 4099)
    st_t = st_t.masked_fill(sq_t == 31, 4100)

    sq_j = jnp.array(sequence_tokens)
    st_j = jnp.array(structure_tokens)
    st_j = jnp.where(st_j == -1, 4096, st_j)
    st_j = jnp.where(sq_j == 0, 4098, st_j)
    st_j = jnp.where(sq_j == 1, 4097, st_j)
    st_j = jnp.where(sq_j == 2, 4099, st_j)
    st_j = jnp.where(sq_j == 31, 4100, st_j)

    print("=== Step 1: Modified structure tokens ===")
    print("Match:", np.array_equal(st_t.numpy(), np.array(st_j)))

    with torch.no_grad():
        torch_enc = (
            torch_model.encoder(
                sq_t,
                st_t,
                torch.tensor(average_plddt),
                torch.from_numpy(per_res_plddt),
                torch.from_numpy(ss8_tokens),
                torch.from_numpy(sasa_tokens),
                torch.from_numpy(function_tokens),
                torch.from_numpy(residue_annotation_tokens),
            )
            .detach()
            .numpy()
        )

    def jax_encode_single(seq, struct, plddt, ss8, sasa, func, residue):
        return jax_model.encoder(
            seq, struct, average_plddt, plddt, ss8, sasa, func, residue
        )

    jax_enc = np.array(
        eqx.filter_vmap(jax_encode_single)(
            sq_j,
            st_j,
            jnp.array(per_res_plddt),
            jnp.array(ss8_tokens),
            jnp.array(sasa_tokens),
            jnp.array(function_tokens),
            jnp.array(residue_annotation_tokens),
        )
    )

    print("=== Step 2: Encoder output ===")
    print("Match:", np.allclose(torch_enc, jax_enc, atol=1e-4))
    print("Max diff:", np.abs(torch_enc - jax_enc).max())

    from esm.utils.structure.affine3d import (
        build_affine3d_from_coordinates as torch_build_affine,
    )

    from jaxonmodels.models.esm import (
        build_affine3d_from_coordinates as jax_build_affine,
    )

    torch_coords = torch.from_numpy(structure_coords)[..., :3, :]
    torch_affine, torch_affine_mask = torch_build_affine(torch_coords)
    torch_affine_tensor = torch_affine.tensor.detach().numpy()
    torch_affine_mask_np = torch_affine_mask.numpy()

    def jax_build_single(c):
        affine, mask = jax_build_affine(c)
        return affine.tensor, mask

    jax_affine_tensor, jax_affine_mask = eqx.filter_vmap(jax_build_single)(
        jnp.array(structure_coords)[..., :3, :]
    )

    print("=== Step 3: Affine ===")
    print(
        "Mask match:", np.array_equal(torch_affine_mask_np, np.array(jax_affine_mask))
    )
    print(
        "Tensor match:",
        np.allclose(
            torch_affine_tensor, np.array(jax_affine_tensor), atol=1e-5, equal_nan=True
        ),
    )
    print(
        "Affine max diff:",
        np.nanmax(np.abs(torch_affine_tensor - np.array(jax_affine_tensor))),
    )

    with torch.no_grad():
        torch_x = torch.from_numpy(torch_enc)
        torch_trans_out, torch_embed, torch_hiddens = torch_model.transformer(
            torch_x,
            None,
            torch_affine,
            torch_affine_mask,
            torch.from_numpy(chain_id),
        )

    from jaxonmodels.models.esm import Affine3D as JaxAffine3D

    def jax_transformer_single(x_i, affine_t_i, affine_m_i, chain_id_i):
        affine_i = JaxAffine3D.from_tensor(affine_t_i)
        return jax_model.transformer(x_i, None, affine_i, affine_m_i, chain_id_i)

    jax_trans_out, jax_embed, jax_hiddens = eqx.filter_vmap(
        jax_transformer_single, in_axes=(0, 0, 0, 0)
    )(
        jnp.array(torch_enc),
        jax_affine_tensor,
        jax_affine_mask,
        jnp.array(chain_id),
    )

    print("=== Step 4: Transformer output ===")
    print(
        "Match:",
        np.allclose(
            torch_trans_out.detach().numpy(), np.array(jax_trans_out), atol=1e-3
        ),
    )
    print(
        "Max diff:",
        np.abs(torch_trans_out.detach().numpy() - np.array(jax_trans_out)).max(),
    )

    for i, (th, jh) in enumerate(zip(torch_hiddens, jax_hiddens)):
        diff = np.abs(th.detach().numpy() - np.array(jh)).max()
        print(f"Hidden layer {i} max diff: {diff}")


def test_esm3_forward_no_vmap():
    d_model, n_heads, v_heads, n_layers, seq_len = 64, 4, 4, 2, 7
    np.random.seed(42)

    sequence_tokens = np.random.randint(3, 31, size=(1, seq_len)).astype(np.int64)
    sequence_tokens[:, 0] = 0
    sequence_tokens[:, -1] = 2
    structure_tokens = np.random.randint(0, 4096, size=(1, seq_len)).astype(np.int64)
    ss8_tokens = np.full((1, seq_len), 8, dtype=np.int64)
    sasa_tokens = np.full((1, seq_len), 16, dtype=np.int64)
    function_tokens = np.full((1, seq_len, 8), 0, dtype=np.int64)
    residue_annotation_tokens = np.full((1, seq_len, 16), 0, dtype=np.int64)
    average_plddt = np.float32(1.0)
    per_res_plddt = np.zeros((1, seq_len), dtype=np.float32)
    structure_coords = np.full((1, seq_len, 3, 3), np.nan, dtype=np.float32)
    chain_id = np.zeros((1, seq_len), dtype=np.int64)

    torch_model = TorchESM3(
        d_model,
        n_heads,
        v_heads,
        n_layers,
        structure_encoder_fn=None,  # ty: ignore[invalid-argument-type]
        structure_decoder_fn=None,  # ty: ignore[invalid-argument-type]
        function_decoder_fn=None,  # ty: ignore[invalid-argument-type]
        tokenizers=_MockTokenizers(),  # ty: ignore[invalid-argument-type]
    )
    state_dict = torch_model.state_dict()

    with torch.no_grad():
        torch_out = torch_model(
            sequence_tokens=torch.from_numpy(sequence_tokens),
            structure_tokens=torch.from_numpy(structure_tokens),
            ss8_tokens=torch.from_numpy(ss8_tokens),
            sasa_tokens=torch.from_numpy(sasa_tokens),
            function_tokens=torch.from_numpy(function_tokens),
            residue_annotation_tokens=torch.from_numpy(residue_annotation_tokens),
            average_plddt=torch.tensor(average_plddt),
            per_res_plddt=torch.from_numpy(per_res_plddt),
            structure_coords=torch.from_numpy(structure_coords),
            chain_id=torch.from_numpy(chain_id),
        )
    torch_seq_logits = torch_out.sequence_logits.numpy()

    key = jax.random.key(42)
    jax_model = JaxESM3(d_model, n_heads, v_heads, n_layers, key=key)
    jax_model = autoconvert(jax_model, state_dict)

    jax_out = jax_model(
        sequence_tokens=jnp.array(sequence_tokens[0]),
        structure_tokens=jnp.array(structure_tokens[0]),
        ss8_tokens=jnp.array(ss8_tokens[0]),
        sasa_tokens=jnp.array(sasa_tokens[0]),
        function_tokens=jnp.array(function_tokens[0]),
        residue_annotation_tokens=jnp.array(residue_annotation_tokens[0]),
        average_plddt=average_plddt,
        per_res_plddt=jnp.array(per_res_plddt[0]),
        structure_coords=jnp.array(structure_coords[0]),
        chain_id=jnp.array(chain_id[0]),
    )

    jax_seq_logits = np.array(jax_out[0])

    print("=== No-vmap test ===")
    print("Match:", np.allclose(torch_seq_logits[0], jax_seq_logits, atol=1e-3))
    print("Max diff:", np.abs(torch_seq_logits[0] - jax_seq_logits).max())
    print("Has NaN (torch):", np.any(np.isnan(torch_seq_logits)))
    print("Has NaN (jax):", np.any(np.isnan(jax_seq_logits)))

    assert np.allclose(torch_seq_logits[0], jax_seq_logits, atol=1e-3)


@pytest.mark.skip("This test takes a while, so local test is sufficient")
@pytest.mark.parametrize("model_name", ["esmc_300m", "esmc_600m"])
def test_full_model_esmc(model_name):

    jax_esmc_300 = JaxESMC.from_pretrained(model_name)

    from esm.models.esmc import ESMC
    from esm.sdk.api import (
        ESMProtein,
    )

    protein = ESMProtein(sequence="AAAAA")
    client = ESMC.from_pretrained(model_name)
    protein_tensor = client.encode(protein)

    sequence_tokens = protein_tensor.sequence.detach().numpy()
    jax_logits, jax_embeddings, jax_hidden = jax_esmc_300(sequence_tokens)
    torch_output = client.forward(protein_tensor.sequence.unsqueeze(0))

    torch_logits = torch_output.sequence_logits.squeeze(0).detach().numpy()
    assert np.allclose(
        np.array(jax_logits),
        torch_logits,
        atol=9e-3,
    )
    torch_embeddings = torch_output.embeddings.squeeze(0).detach().numpy()
    assert np.allclose(
        np.array(jax_embeddings),
        torch_embeddings,
        atol=9e-3,
    )
    torch_hidden = torch_output.hidden_states.squeeze(1).detach().numpy()

    assert np.allclose(
        np.array(jax_hidden),
        torch_hidden,
        atol=9e-3,
    )


@pytest.mark.skip("Too long for CI/CD")
def test_full_model_esm3():
    jax_esm3 = JaxESM3.from_pretrained("esm3_open")

    from esm.models.esm3 import ESM3 as TorchESM3
    from esm.sdk.api import ESMProtein

    client = TorchESM3.from_pretrained("esm3-open")

    protein = ESMProtein(sequence="AAAAA")
    protein_tensor = client.encode(protein)

    sequence_tokens = protein_tensor.sequence

    with torch.no_grad():
        torch_out = client.forward(
            sequence_tokens=sequence_tokens.unsqueeze(0),
        )

    jax_out = jax_esm3(
        sequence_tokens=jnp.array(sequence_tokens.numpy()),
    )

    torch_seq_logits = torch_out.sequence_logits.squeeze(0).numpy()
    jax_seq_logits = np.array(jax_out[0])

    print(f"{torch_seq_logits.shape=}")
    print(f"{jax_seq_logits.shape=}")
    print(f"max diff: {np.abs(torch_seq_logits - jax_seq_logits).max()}")

    assert np.allclose(jax_seq_logits, torch_seq_logits, atol=9e-3)


@pytest.mark.parametrize("d_model, seq_len", [(64, 7), (128, 11)])
def test_dim6rot_structure_head(d_model, seq_len):
    np.random.seed(42)

    x = np.random.randn(1, seq_len, d_model).astype(np.float32)
    affine_mask = np.zeros((1, seq_len), dtype=np.bool_)

    from esm.layers.structure_proj import Dim6RotStructureHead as TorchDim6Rot

    torch_head = TorchDim6Rot(d_model, trans_scale_factor=10)
    state_dict = torch_head.state_dict()

    with torch.no_grad():
        torch_affine, torch_xyz = torch_head(
            torch.from_numpy(x),
            affine=None,
            affine_mask=torch.from_numpy(affine_mask),
        )
    torch_affine = torch_affine.numpy()
    torch_xyz = torch_xyz.numpy()

    key = jax.random.key(42)
    jax_head = JaxDim6Rot(d_model, trans_scale_factor=10, key=key)
    jax_head = autoconvert(jax_head, state_dict)

    jax_affine, jax_xyz = jax_head(
        jnp.array(x[0]),
        affine=None,
        affine_mask=jnp.array(affine_mask[0]),
    )

    print(f"affine max diff: {np.abs(torch_affine[0] - np.array(jax_affine)).max()}")
    print(f"xyz max diff: {np.abs(torch_xyz[0] - np.array(jax_xyz)).max()}")

    assert np.allclose(np.array(jax_affine), torch_affine[0], atol=5e-3)
    assert np.allclose(np.array(jax_xyz), torch_xyz[0], atol=5e-3)


@pytest.mark.parametrize("d_model, seq_len", [(64, 7), (128, 11)])
def test_pairwise_prediction_head(d_model, seq_len):
    np.random.seed(42)

    x = np.random.randn(1, seq_len, d_model).astype(np.float32)

    from esm.models.vqvae import PairwisePredictionHead as TorchPairwise

    torch_head = TorchPairwise(
        input_dim=d_model,
        downproject_dim=128,
        hidden_dim=128,
        n_bins=224,
        bias=False,
    )
    state_dict = torch_head.state_dict()

    with torch.no_grad():
        torch_out = torch_head(torch.from_numpy(x)).numpy()

    key = jax.random.key(42)
    jax_head = JaxPairwisePredictionHead(
        input_dim=d_model,
        downproject_dim=128,
        hidden_dim=128,
        n_bins=224,
        bias=False,
        key=key,
    )
    jax_head = autoconvert(jax_head, state_dict)

    jax_out = np.array(jax_head(jnp.array(x[0])))

    print(f"shape torch: {torch_out.shape}")
    print(f"shape jax: {jax_out.shape}")
    print(f"max diff: {np.abs(torch_out[0] - jax_out).max()}")

    assert np.allclose(jax_out, torch_out[0], atol=1e-3)


@pytest.mark.parametrize(
    "d_model, n_heads, n_layers, seq_len",
    [
        (64, 4, 2, 7),
        (128, 8, 2, 11),
    ],
)
def test_structure_token_decoder(d_model, n_heads, n_layers, seq_len):
    np.random.seed(42)

    structure_tokens = np.random.randint(0, 4096, size=(1, seq_len)).astype(np.int64)
    structure_tokens[:, 0] = VQVAE_SPECIAL_TOKENS["BOS"]
    structure_tokens[:, -1] = VQVAE_SPECIAL_TOKENS["EOS"]

    from esm.models.vqvae import StructureTokenDecoder as TorchDecoder

    torch_decoder = TorchDecoder(d_model, n_heads, n_layers)
    state_dict = torch_decoder.state_dict()

    with torch.no_grad():
        torch_out = torch_decoder.decode(torch.from_numpy(structure_tokens))

    torch_affine = torch_out["tensor7_affine"].numpy()
    torch_bb = torch_out["bb_pred"].numpy()
    torch_plddt = torch_out["plddt"].numpy()
    torch_ptm = torch_out["ptm"].numpy()
    torch_pae = torch_out["predicted_aligned_error"].numpy()

    key = jax.random.key(42)
    jax_decoder = JaxStructureTokenDecoder(d_model, n_heads, n_layers, key=key)
    jax_decoder = autoconvert(jax_decoder, state_dict)

    jax_out = jax_decoder.decode(jnp.array(structure_tokens[0]))

    jax_affine = np.array(jax_out["tensor7_affine"])
    jax_bb = np.array(jax_out["bb_pred"])
    jax_plddt = np.array(jax_out["plddt"])
    jax_ptm = np.array(jax_out["ptm"])
    jax_pae = np.array(jax_out["predicted_aligned_error"])

    print(f"affine max diff: {np.abs(torch_affine[0] - jax_affine).max()}")
    print(f"bb max diff: {np.abs(torch_bb[0] - jax_bb).max()}")
    print(f"plddt max diff: {np.abs(torch_plddt[0] - jax_plddt).max()}")
    print(f"ptm diff: {np.abs(torch_ptm.item() - jax_ptm.item())}")
    print(f"pae max diff: {np.abs(torch_pae[0] - jax_pae).max()}")

    assert np.allclose(jax_affine, torch_affine[0], atol=5e-3)
    assert np.allclose(jax_bb, torch_bb[0], atol=5e-3)
    assert np.allclose(jax_plddt, torch_plddt[0], atol=1e-3)
    assert np.abs(torch_ptm.item() - jax_ptm.item()) < 1e-3
    assert np.allclose(jax_pae, torch_pae[0], atol=1e-3)


@pytest.mark.skip("Because we get huggingface_hub.errors.GatedRepoError")
def test_structure_token_decoder_pretrained():
    from esm.pretrained import ESM3_structure_decoder_v0

    torch_decoder = ESM3_structure_decoder_v0("cpu")
    jax_decoder = JaxStructureTokenDecoder.from_pretrained()

    np.random.seed(42)
    seq_len = 7
    structure_tokens = np.random.randint(0, 4096, size=(1, seq_len)).astype(np.int64)
    structure_tokens[:, 0] = VQVAE_SPECIAL_TOKENS["BOS"]
    structure_tokens[:, -1] = VQVAE_SPECIAL_TOKENS["EOS"]

    with torch.no_grad():
        torch_out = torch_decoder.decode(torch.from_numpy(structure_tokens))

    jax_out = jax_decoder.decode(jnp.array(structure_tokens[0]))

    torch_bb = torch_out["bb_pred"].numpy()
    jax_bb = np.array(jax_out["bb_pred"])
    torch_plddt = torch_out["plddt"].numpy()
    jax_plddt = np.array(jax_out["plddt"])
    torch_ptm = torch_out["ptm"].numpy()
    jax_ptm = np.array(jax_out["ptm"])

    print(f"bb max diff: {np.abs(torch_bb[0] - jax_bb).max()}")
    print(f"plddt max diff: {np.abs(torch_plddt[0] - jax_plddt).max()}")
    print(f"ptm diff: {np.abs(torch_ptm.item() - jax_ptm.item())}")

    assert np.allclose(jax_bb, torch_bb[0], atol=0.5)
    assert np.allclose(jax_plddt, torch_plddt[0], atol=5e-2)
    assert np.abs(torch_ptm.item() - jax_ptm.item()) < 5e-2
